(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2010 Xavier Clerc.
 * Original file (driver/opterrors.ml in the Objective Caml source
 * distribution) is Copyright (C) INRIA.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://qt.nokia.com/doc/4.0/qpl.html>.
 *)

open Format

(* Report an error *)

let report_error ppf exn =
  let barista_error msg =
    fprintf ppf "Barista error: %s" msg in
  let report ppf = function
  | Lexer.Error(err, l) ->
      Location.print_error ppf l;
      Lexer.report_error ppf err
  | Syntaxerr.Error err ->
      Syntaxerr.report_error ppf err
  | Pparse.Error ->
      Location.print_error_cur_file ppf;
      fprintf ppf "Preprocessor error"
  | Env.Error err ->
      Location.print_error_cur_file ppf;
      Env.report_error ppf err
  | Ctype.Tags(l, l') ->
      Location.print_error_cur_file ppf;
      fprintf ppf
      "In this program,@ variant constructors@ `%s and `%s@ \
       have the same hash value.@ Change one of them." l l'
  | Typecore.Error(loc, err) ->
      Location.print_error ppf loc; Typecore.report_error ppf err
  | Typetexp.Error(loc, err) ->
      Location.print_error ppf loc; Typetexp.report_error ppf err
  | Typedecl.Error(loc, err) ->
      Location.print_error ppf loc; Typedecl.report_error ppf err
  | Typeclass.Error(loc, err) ->
      Location.print_error ppf loc; Typeclass.report_error ppf err
  | Includemod.Error err ->
      Location.print_error_cur_file ppf;
      Includemod.report_error ppf err
  | Typemod.Error(loc, err) ->
      Location.print_error ppf loc; Typemod.report_error ppf err
  | Translcore.Error(loc, err) ->
      Location.print_error ppf loc; Translcore.report_error ppf err
  | Translclass.Error(loc, err) ->
      Location.print_error ppf loc; Translclass.report_error ppf err
  | Translmod.Error(loc, err) ->
      Location.print_error ppf loc; Translmod.report_error ppf err
  | Jcompilenv.Error code ->
      Location.print_error_cur_file ppf;
      Jcompilenv.report_error ppf code
  | Bytecodegen.Error code ->
      Location.print_error_cur_file ppf;
      Bytecodegen.report_error ppf code
  | Cadmiumprimitives.Error code ->
      Location.print_error_cur_file ppf;
      Cadmiumprimitives.report_error ppf code
  | Javalink.Error code ->
      Location.print_error_cur_file ppf;
      Javalink.report_error ppf code
  | Javalibrarian.Error code ->
      Location.print_error_cur_file ppf;
      Javalibrarian.report_error ppf code
  | Javapackager.Error code ->
      Location.print_error_cur_file ppf;
      Javapackager.report_error ppf code
  | BaristaLibrary.AccessFlag.Exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.AccessFlag.string_of_error err)
  | BaristaLibrary.Annotation.Exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.Annotation.string_of_error err)
  | BaristaLibrary.Attribute.Exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.Attribute.string_of_error err)
  | BaristaLibrary.ByteCode.Exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.ByteCode.string_of_error err)
  | BaristaLibrary.ClassDefinition.Exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.ClassDefinition.string_of_error err)
  | BaristaLibrary.ClassFile.Exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.ClassFile.string_of_error err)
  | BaristaLibrary.ClassLoader.Exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.ClassLoader.string_of_error err)
  | BaristaLibrary.ClassPath.Exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.ClassPath.string_of_error err)
  | BaristaLibrary.ConstantPool.Exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.ConstantPool.string_of_error err)
  | BaristaLibrary.Descriptor.Exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.Descriptor.string_of_error err)
  | BaristaLibrary.Field.Exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.Field.string_of_error err)
  | BaristaLibrary.InputStream.Exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.InputStream.string_of_error err)
  | BaristaLibrary.Instruction.Exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.Instruction.string_of_error err)
  | BaristaLibrary.Method.Exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.Method.string_of_error err)
  | BaristaLibrary.Name.Exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.Name.string_of_error err)
  | BaristaLibrary.OutputStream.Exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.OutputStream.string_of_error err)
  | BaristaLibrary.Signature.Exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.Signature.string_of_error err)
  | BaristaLibrary.Utils.Unicode_exception err ->
      Location.print_error_cur_file ppf;
      barista_error (BaristaLibrary.Utils.string_of_unicode_error err)
  | Sys_error msg ->
      Location.print_error_cur_file ppf;
      fprintf ppf "I/O error: %s" msg
  | Warnings.Errors (n) ->
      Location.print_error_cur_file ppf;
      fprintf ppf "Error: error-enabled warnings (%d occurrences)" n
  | x -> fprintf ppf "@]"; raise x in
  fprintf ppf "@[%a@]@." report exn
