(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://qt.nokia.com/doc/4.0/qpl.html>.
 *)

open Macroinstr
open Bytecodeutils
open Instruct
open Lambda
open Asttypes
open BaristaLibrary
open Utils


type error =
  | Method_too_long of string
  | Unable_to_compute_stack of string

exception Error of error


(* compilation state *)

let current_file = ref ""

let current_class = ref ""

let current_function = ref ""

let exception_table = ref []

let garbage_index = ref 0

let current_catches = Hashtbl.create 17

let line_number_table = ref []

let prev_ofs = ref 0


(* constants *)

let constants = ref []

let reset_constants () =
  constants := []

let add_constant c =
  let res = List.length !constants in
  constants := c :: !constants;
  res

let compile_constants () =
  let rec conv = function
    | Const_base (Const_int i) -> Obj.repr i
    | Const_base (Const_char c) -> Obj.repr c
    | Const_base (Const_string s) -> Obj.repr s
    | Const_base (Const_float s) -> Obj.repr (float_of_string s)
    | Const_base (Const_int32 i) -> Obj.repr i
    | Const_base (Const_int64 i) -> Obj.repr i
    | Const_base (Const_nativeint i) -> Obj.repr i
    | Const_pointer i -> Obj.repr i
    | Const_immstring s -> Obj.repr s
    | Const_block (tag, fields) ->
        let block = Obj.new_block tag (List.length fields) in
        let pos = ref 0 in
        List.iter
          (fun c ->
            Obj.set_field block !pos (conv c);
            incr pos)
          fields;
        block
    | Const_float_array fields ->
        Obj.repr (Array.of_list (List.map float_of_string fields)) in
  let towrite = List.rev_map conv !constants in
  let array = Array.make (List.length towrite) (Obj.magic 0) in
  let i = ref 0 in
  List.iter
    (fun c ->
      array.(!i) <- c;
      incr i)
    towrite;
  Marshal.to_string array []

let compile_constant boxed_result c =
  let make_long x =
    match boxed_result with
    | Boxed -> push_boxed_int x
    | Unboxed -> push_int x in
  match c with
  | Const_base (Const_int i)
  | Const_pointer i ->
      make_long i
  | Const_base (Const_char c) ->
      make_long (Char.code c)
  | _ ->
      let idx = add_constant c in
      let idx_index, idx_index_sz = push_int idx in
      [ Instruction.ALOAD_0 ]
      @ [ Instruction.LDC_W (`Class_or_interface (make_class !current_class)) ]
      @ idx_index
      @ meth_getConstant
      @ (if boxed_result = Unboxed then meth_asLong else []),
      1 + 3 + idx_index_sz + meth_size
        + (if boxed_result = Unboxed then meth_size else 0)


(* expression compilation *)

let parameters n =
  let rec params acc = function
    | 0 -> acc
    | n -> params ((`Class class_Value) :: acc) (pred n) in
  params [] n

let check_jump ofs =
  if ofs >= -32768 && ofs <= 32767 then
    ofs
  else
    raise (Error (Method_too_long !current_function))

let add_debug_info ofs dbg =
  if dbg != Debuginfo.none && dbg.Debuginfo.dinfo_file = !current_file then
    begin
      line_number_table :=
        (!prev_ofs, dbg.Debuginfo.dinfo_line) :: !line_number_table;
      prev_ofs := ofs
    end

let take n l =
  let rec take' n acc l =
    match n, l with
    | 0, _ -> acc
    | _, hd :: tl -> take' (pred n) (hd :: acc) tl
    | _ -> assert false in
  List.rev (take' n [] l)

let rec compile_expression stack is_tail boxed_result ofs = function
  | Mconst c ->
      compile_constant boxed_result c
  | Moffset (expr, n) ->
      let expr_instrs, expr_sz =
        compile_expression stack false Boxed ofs expr in
      let int_instrs, int_sz = push_int n in
      expr_instrs
      @ meth_asBlock
      @ int_instrs
      @ meth_offset
      @ (if boxed_result = Unboxed then meth_asLong else []),
      expr_sz + meth_size + int_sz + meth_size
        + (if boxed_result = Unboxed then meth_size else 0)
  | Mdynamicoffset (expr, n) ->
      let expr_instrs, expr_sz =
        compile_expression stack false Boxed ofs expr in
      let int_instrs, int_sz =
        compile_expression (Boxed :: stack) false Unboxed (ofs + expr_sz + meth_size + 1 + meth_size + meth_size) n in
      expr_instrs
      @ meth_asBlock
      @ [ Instruction.ICONST_0 ]
      @ meth_get
      @ meth_asBlock
      @ int_instrs
      @ meth_get
      @ (if boxed_result = Unboxed then meth_asLong else []),
      expr_sz + meth_size + 1 + meth_size + meth_size + int_sz + meth_size
        + (if boxed_result = Unboxed then meth_size else 0)
  | Mcreateglobal (id, sz) ->
      let size_instrs, size_sz = push_int sz in
      [ Instruction.ALOAD_0 ;
        Instruction.LDC_W (`String (utf8_of_string id)) ]
      @ size_instrs
      @ meth_createGlobal,
      1 + 3 + size_sz + meth_size
  | Mclosure (2, ((cls, func), 1) :: [], []) ->
      [ Instruction.ALOAD_0 ;
        Instruction.LDC_W (`Class_or_interface (make_class cls)) ;
        Instruction.LDC_W (`String (normalize_function_name func)) ]
      @ meth_registerClosure1,
      1 + 3 + 3 + meth_size
  | Mclosure (3, ((cls, func), arity) :: [], []) ->
      let arity_instrs, arity_sz = push_int arity in
      [ Instruction.ALOAD_0 ;
        Instruction.LDC_W (`Class_or_interface (make_class cls)) ;
        Instruction.LDC_W (`String (normalize_function_name func)) ]
      @ arity_instrs
      @ meth_registerClosureN,
      1 + 3 + 3 + arity_sz + meth_size
  | Mclosure (sz, ((cls, func), arity) :: others, vars) ->
      let size_instrs, size_sz = push_int sz in
      let load_instrs, load_sz = aload !garbage_index in
      let store_instrs, store_sz = astore !garbage_index in
      let first_instrs, first_sz, next =
        compile_first_closure cls func arity in
      let others_instrs, others_sz, index =
        compile_other_closures next others in
      let vars_instrs, vars_sz =
        compile_expression_list (Boxed :: stack) Boxed (ofs + size_sz + meth_size + 1 + 1 + meth_size + store_sz + first_sz + others_sz + load_sz) vars in
      let index_instrs, index_sz = push_int index in
      let n_instrs, n_sz = push_int (List.length vars) in
      let fold_instrs = ref [] in
      let fold_sz = ref 0 in
      let remaining = ref (List.length vars) in
      while !remaining > 0 do
        let n = min 4 !remaining in
        fold_instrs := !fold_instrs
          @ [ Instruction.INVOKESTATIC
                (class_AbstractNativeRunner,
                 make_method ("foldClosure" ^ (string_of_int n)),
                 ((parameters n) @ [`Class class_Closure],
                  `Class class_Closure)) ];
        fold_sz := !fold_sz + meth_size;
        remaining := !remaining - n
      done;
      size_instrs
      @ meth_createClosure
      @ [ Instruction.DUP ;
          Instruction.DUP ]
      @ meth_setCustom
      @ store_instrs
      @ first_instrs
      @ others_instrs
      @ load_instrs
      @ vars_instrs
      @ index_instrs
      @ n_instrs
      @ meth_createClosureVars
      @ !fold_instrs
      @ meth_returnClosure,
      size_sz + meth_size + 1 + 1 + meth_size + store_sz + first_sz + others_sz
        + load_sz + vars_sz + index_sz + n_sz + meth_size + !fold_sz + meth_size
  | Mclosure _ ->
      Misc.fatal_error "Bytecodegen.compile_expression: invalid closure"
  | Mreadlocal idx ->
      if boxed_result = Boxed then
        aload idx
      else
        let aload_instrs, aload_sz = aload idx in
        aload_instrs
        @ meth_asLong,
        aload_sz + meth_size
  | Mreadlocalint idx ->
      if boxed_result = Unboxed then
        iload idx
      else
        let iload_instrs, iload_sz = iload idx in
        iload_instrs
        @ meth_createFromLong,
        iload_sz + meth_size
  | Mwritelocal (idx, expr) ->
      let expr_instrs, expr_sz =
        compile_expression stack false Boxed ofs expr in
      let astore_instrs, astore_sz = astore idx in
      expr_instrs
      @ astore_instrs,
      expr_sz + astore_sz
  | Mwritelocalint (idx, expr) ->
      let expr_instrs, expr_sz =
        compile_expression stack false Unboxed ofs expr in
      let istore_instrs, istore_sz = istore idx in
      expr_instrs
      @ istore_instrs,
      expr_sz + istore_sz
  | Mcall (call, (cls, func), args, dbg) ->
      add_debug_info ofs dbg;
      if is_tail && cls = !current_class && func = !current_function then
        let args_instrs, args_sz =
          compile_expression_list_to_locals (Boxed :: stack) (succ ofs) args in
        [ Instruction.ACONST_NULL ]
        @ args_instrs
        @ [ Instruction.POP ]
        @ [ Instruction.GOTO (check_jump ~-(ofs + 1 + args_sz + 1)) ],
        1 + args_sz + 1 + 3
      else begin
        let arity = List.length args in
        let args_instrs, args_sz =
          compile_expression_list (Boxed :: stack) Boxed (succ ofs) args in
        [ Instruction.ALOAD_0 ]
        @ args_instrs
        @ [ Instruction.INVOKESTATIC
              (make_class cls,
               make_method func,
               ((`Class (if call = Native then class_AbstractNativeRunner else class_CodeRunner))
                :: (parameters arity), `Class class_Value)) ]
        @ (if boxed_result = Unboxed then meth_asLong else []),
        1 + args_sz + meth_size
          + (if boxed_result = Unboxed then meth_size else 0)
      end
  | Mprim (prim, args, dbg) ->
      add_debug_info ofs dbg;
      (try
        compile_expression stack false boxed_result ofs (Bytecodeprimitive.as_mcall prim args)
      with
      | Bytecodeprimitive.Not_a_call ->
          match prim, args with
          | Pidentity, [arg] ->
              compile_expression stack false boxed_result ofs arg
          | Pignore, [arg] ->
              let arg_instrs, arg_sz =
                compile_expression stack false Boxed ofs arg in
              arg_instrs
              @ [ Instruction.POP ]
              @ field_UNIT,
              arg_sz + 1 + field_size
          | Pnot, [arg] ->
              (match boxed_result with
              | Boxed ->
                  let arg_instrs, arg_sz =
                    compile_expression stack false Boxed ofs arg in
                  arg_instrs
                  @ field_FALSE
                  @ [ Instruction.IF_ACMPEQ (3 + field_size + 3) ]
                  @ field_FALSE
                  @ [ Instruction.GOTO (3 + field_size) ]
                  @ field_TRUE,
                  arg_sz + field_size + 3 + field_size + 3 + field_size
              | Unboxed ->
                  let arg_instrs, arg_sz =
                    compile_expression stack false Unboxed ofs arg in
                  arg_instrs
                  @ [ Instruction.IFEQ (3 + 1 + 3) ;
                      Instruction.ICONST_0 ;
                      Instruction.GOTO (3 + 1) ;
                      Instruction.ICONST_1 ],
                  arg_sz + 3 + 1 + 3 + 1)
          | Psequand, [expr1; expr2] ->
              let expr1_instrs, expr1_sz =
                compile_expression stack false Unboxed ofs expr1 in
              let expr2_instrs, expr2_sz =
                compile_expression stack false Unboxed (ofs + expr1_sz + 3) expr2 in
              expr1_instrs
              @ [ Instruction.IFEQ (check_jump (3 + expr2_sz + 3)) ]
              @ expr2_instrs
              @ [ Instruction.GOTO (3 + 1) ;
                  Instruction.ICONST_0 ]
              @ (if boxed_result = Boxed then meth_createFromLong else []),
              expr1_sz + 3 + expr2_sz + 3 + 1
                + (if boxed_result = Boxed then meth_size else 0)
          | Psequor, [expr1; expr2] ->
              let expr1_instrs, expr1_sz =
                compile_expression stack false Unboxed ofs expr1 in
              let expr2_instrs, expr2_sz =
                compile_expression stack false Unboxed (ofs + expr1_sz + 3) expr2 in
              expr1_instrs
              @ [ Instruction.IFNE (check_jump (3 + expr2_sz + 3)) ]
              @ expr2_instrs
              @ [ Instruction.GOTO (3 + 1) ;
                  Instruction.ICONST_1 ]
              @ (if boxed_result = Boxed then meth_createFromLong else []),
              expr1_sz + 3 + expr2_sz + 3 + 1
                + (if boxed_result = Boxed then meth_size else 0)
          | Pstringrefu, [str; idx] ->
              let str_instrs, str_sz =
                compile_expression stack false Boxed ofs str in
              let idx_instrs, idx_sz =
                compile_expression (Boxed :: stack) false Unboxed (ofs + str_sz + meth_size) idx in
              str_instrs
              @ meth_asBlock
              @ idx_instrs
              @ meth_getUnsignedByte
              @ (if boxed_result = Boxed then meth_createFromLong else []),
              str_sz + meth_size + idx_sz + meth_size
                + (if boxed_result = Boxed then meth_size else 0)
          | Pstringsetu, [str; idx; v] ->
              let str_instrs, str_sz =
                compile_expression stack false Boxed ofs str in
              let idx_instrs, idx_sz =
                compile_expression (Boxed :: stack) false Unboxed (ofs + str_sz + meth_size) idx in
              let v_instrs, v_sz =
                compile_expression (Unboxed :: Boxed :: stack) false Unboxed (ofs + str_sz + meth_size + idx_sz) v in
              str_instrs
              @ meth_asBlock
              @ idx_instrs
              @ v_instrs
              @ [ Instruction.SIPUSH 0x00FF ;
                  Instruction.IAND ]
              @ meth_setUnsignedByte
              @ field_UNIT,
              str_sz + meth_size + idx_sz + v_sz + 3 + 1 + meth_size + field_size
          | (Pmakearray _), [] ->
              compile_constant Boxed (Const_block (0, []))
          | (Pmakearray kind), args ->
              (match kind with
              | Pintarray | Paddrarray ->
                  compile_block stack 0 Mutable ofs args
              | Pfloatarray ->
                  let size = List.length args in
                  let args_instrs, args_sz =
                    compile_expression_list stack Boxed ofs args in
                  let size_instrs, size_sz = push_int size in
                  let fold_instrs = ref [] in
                  let fold_sz = ref 0 in
                  let remaining = ref size in
                  while !remaining > 0 do
                    let n = min 4 !remaining in
                    fold_instrs := !fold_instrs
                      @ [ Instruction.INVOKESTATIC
                            (class_AbstractNativeRunner,
                             make_method ("foldValues" ^ (string_of_int n)),
                             ((parameters n) @ [`Class class_Values],
                              `Class class_Values)) ];
                    fold_sz := !fold_sz + meth_size;
                    remaining := !remaining - n
                  done;
                  args_instrs
                  @ size_instrs
                  @ meth_createDoubles
                  @ !fold_instrs
                  @ meth_returnDoubles,
                  args_sz + size_sz + meth_size + !fold_sz + meth_size
              | Pgenarray ->
                  let block_instrs, block_sz =
                    compile_block stack 0 Mutable ofs args in
                  block_instrs
                  @ [ Instruction.ALOAD_0 ;
                      Instruction.SWAP ]
                  @ meth_caml_make_array,
                  block_sz + 1 + 1 + meth_size)
          | (Parrayrefu _), [arr; idx] ->
              let arr_instrs, arr_sz =
                compile_expression stack false Boxed ofs arr in
              let idx_instrs, idx_sz =
                compile_expression (Boxed :: stack) false Unboxed (ofs + arr_sz + meth_size) idx in
              arr_instrs
              @ meth_asBlock
              @ idx_instrs
              @ meth_get
              @ (if boxed_result = Unboxed then meth_asLong else []),
              arr_sz + meth_size + idx_sz + meth_size + (if boxed_result = Unboxed then meth_size else 0)
          | (Parraysetu _), [arr; idx; v] ->
              let arr_instrs, arr_sz =
                compile_expression stack false Boxed ofs arr in
              let idx_instrs, idx_sz =
                compile_expression (Boxed :: stack) false Unboxed (ofs + arr_sz + meth_size) idx in
              let v_instrs, v_sz =
                compile_expression (Unboxed :: Boxed :: stack) false Boxed (ofs + arr_sz + meth_size + idx_sz) v in
              arr_instrs
              @ meth_asBlock
              @ idx_instrs
              @ v_instrs
              @ meth_set
              @ field_UNIT,
              arr_sz + meth_size + idx_sz + v_sz + meth_size + field_size
          | (Pmakeblock (tag, mut)), args ->
              compile_block stack tag mut ofs args
          | Pisint, [arg] ->
              let arg_instrs, arg_sz =
                compile_expression stack false Boxed ofs arg in
              (match boxed_result with
              | Boxed ->
                  arg_instrs
                  @ meth_isLong
                  @ meth_createFromLong,
                  arg_sz + meth_size + meth_size
              | Unboxed ->
                  arg_instrs
                  @ meth_isLong,
                  arg_sz + meth_size)
          | (Pintcomp cmp), [arg1; arg2] ->
              let arg1_instrs, arg1_sz =
                compile_expression stack false Boxed ofs arg1 in
              let arg2_instrs, arg2_sz =
                compile_expression (Boxed :: stack) false Boxed (ofs + arg1_sz) arg2 in
              arg1_instrs
              @ arg2_instrs
              @ (match cmp with
              | Ceq -> meth_equalValues
              | Cneq -> meth_notEqualValues
              | Clt -> meth_lowerThanValue
              | Cgt -> meth_greaterThanValue
              | Cle -> meth_lowerEqualValue
              | Cge -> meth_greaterEqualValue)
              @ (if boxed_result = Boxed then meth_createFromLong else []),
              arg1_sz + arg2_sz + meth_size
                + (if boxed_result = Boxed then meth_size else 0)
          | Pisout, [arg1; arg2] ->
              let arg1_instrs, arg1_sz =
                compile_expression stack false Boxed ofs arg1 in
              let arg2_instrs, arg2_sz =
                compile_expression (Boxed :: stack) false Boxed (ofs + arg1_sz) arg2 in
              arg1_instrs
              @ arg2_instrs
              @ meth_isOutValue
              @ (if boxed_result = Boxed then meth_createFromLong else []),
              arg1_sz + arg2_sz + meth_size
                + (if boxed_result = Boxed then meth_size else 0)
          | _ ->
              let boxed = Bytecodeprimitive.parameters_type prim in
              let args_instrs, args_sz =
                compile_expression_list stack boxed ofs args in
              let prim_instrs, prim_sz =
                Bytecodeprimitive.compile prim in
              let conv_instrs, conv_sz = match boxed, boxed_result with
              | Boxed, Unboxed ->
                  meth_asLong, meth_size
              | Unboxed, Boxed ->
                  meth_createFromLong, meth_size
              | Boxed, Boxed | Unboxed, Unboxed -> [], 0 in
              args_instrs
              @ prim_instrs
              @ conv_instrs,
              args_sz + prim_sz + conv_sz)
  | Mapply (args, dbg) ->
      add_debug_info ofs dbg;
      let arity = pred (List.length args) in
      if (arity <= 4) then begin
        let args_instrs, args_sz =
          compile_expression_list (Boxed :: stack) Boxed (succ ofs) args  in
        [ Instruction.ALOAD_0 ]
        @ args_instrs
        @ [ Instruction.INVOKEVIRTUAL
              ((`Class_or_interface class_AbstractNativeRunner),
               make_method ("apply" ^ (string_of_int (arity))),
               ((parameters (succ arity)), `Class class_Value)) ]
        @ (if boxed_result = Unboxed then meth_asLong else [])
        @ [ Instruction.ALOAD_0 ]
        @ meth_checkSignals,
        1 + args_sz + meth_size + (if boxed_result = Unboxed then meth_size else 0) + 1 + meth_size
      end else begin
        let args_instrs, args_sz =
          compile_expression_list stack Boxed ofs args  in
        let fold_instrs = ref [] in
        let fold_sz = ref 0 in
        let current_arity = ref arity in
        while !current_arity > 0 do
          let reduce = min !current_arity 4 in
          fold_instrs := !fold_instrs
            @ [ Instruction.INVOKESTATIC
                  (class_AbstractNativeRunner,
                   make_method ("foldApplication" ^ (string_of_int reduce)),
                   ((parameters reduce) @ [`Class class_Application], `Class class_Application)) ];
          current_arity := !current_arity - reduce;
          fold_sz := !fold_sz + meth_size
        done;
        let arity_instrs, arity_sz = push_int arity in
        args_instrs
        @ [ Instruction.ALOAD_0 ;
            Instruction.SWAP ]
        @ arity_instrs
        @ meth_createApplication
        @ !fold_instrs
        @ meth_returnApplication
        @ (if boxed_result = Unboxed then meth_asLong else [])
        @ [ Instruction.ALOAD_0 ]
        @ meth_checkSignals,
        args_sz + 2 + arity_sz + meth_size + !fold_sz + meth_size
          + (if boxed_result = Unboxed then meth_size else 0)
          + 1 + meth_size
      end
  | Msequence (e1, e2) ->
      let e1_instrs, e1_sz =
        compile_expression stack false Boxed ofs e1 in
      let e2_instrs, e2_sz =
        compile_expression stack is_tail boxed_result (ofs + e1_sz) e2 in
      e1_instrs
      @ e2_instrs,
      e1_sz + e2_sz
  | Mifthenelse (cond, ifso, ifno) ->
      let cond_instrs, cond_sz =
        compile_expression stack false Unboxed ofs cond in
      let ifso_instrs, ifso_sz =
        compile_expression stack is_tail boxed_result (ofs + cond_sz + 3) ifso in
      let ifno_instrs, ifno_sz =
        compile_expression stack is_tail boxed_result (ofs + cond_sz + 3 + ifso_sz + 3) ifno in
      cond_instrs
      @ [ Instruction.IFEQ (check_jump (3 + ifso_sz + 3)) ]
      @ ifso_instrs
      @ [ Instruction.GOTO (check_jump (3 + ifno_sz)) ]
      @ ifno_instrs,
      cond_sz + 3 + ifso_sz + 3 + ifno_sz
  | Mwhile (cond, body) ->
      let body_instrs, body_sz =
        compile_expression stack false Boxed (ofs + 3 + 1 + meth_size) body in
      let cond_instrs, cond_sz =
        compile_expression stack false Unboxed (ofs + 3 + 1 + meth_size + body_sz) cond in
      [ Instruction.GOTO (check_jump (3 + 1 + meth_size + body_sz)) ;
        Instruction.ALOAD_0 ]
      @ meth_checkSignals
      @ body_instrs
      @ cond_instrs
      @ [ Instruction.IFNE (check_jump ~-(1 + meth_size + body_sz + cond_sz)) ],
      3 + 1 + meth_size + body_sz + cond_sz + 3
  | Mfor (idx, dir, bound, body) ->
      let body_instrs, body_sz =
        compile_expression stack false Boxed (ofs + 3 + 1 + meth_size) body in
      let iinc_instrs, iinc_sz =
        iinc idx (if dir = Upto then 1 else -1) in
      let idx_instrs, idx_sz = iload idx in
      let bound_instrs, bound_sz = iload bound in
      [ Instruction.GOTO (check_jump (3 + 1 + meth_size + body_sz + iinc_sz)) ;
        Instruction.ALOAD_0 ]
      @ meth_checkSignals
      @ body_instrs
      @ iinc_instrs
      @ idx_instrs
      @ bound_instrs
      @ (if dir = Upto then
        [ Instruction.IF_ICMPLE (check_jump ~-(bound_sz + idx_sz + iinc_sz + body_sz + meth_size + 1)) ]
      else
        [ Instruction.IF_ICMPGE (check_jump ~-(bound_sz + idx_sz + iinc_sz + body_sz + meth_size + 1)) ]),
      3 + 1 + meth_size + body_sz + iinc_sz + idx_sz + bound_sz + 3
  | Munit ->
      (match boxed_result with
      | Boxed -> field_UNIT, field_size
      | Unboxed -> push_int 0)
  | Mpoptolocal idx ->
      astore idx
  | Mpop ->
      [ Instruction.POP ], 1
  | Mnop ->
      [], 0
  | Mtrywith (body, idx, handler, backup) ->
      let gload_instrs, gload_sz = aload !garbage_index in
      let gstore_instrs, gstore_sz = astore !garbage_index in
      let save_instrs = ref [] in
      let restore_instrs = ref [] in
      let backup_size = ref 0 in
      assert ((List.length backup) >= (List.length stack));
      List.iter2
        (fun idx kind ->
          let sav, sz =
            match kind with Boxed -> astore idx | Unboxed -> istore idx in
          let rst, _ =
            match kind with Boxed -> aload idx | Unboxed -> iload idx in
          save_instrs := sav :: !save_instrs;
          restore_instrs := rst :: !restore_instrs;
          backup_size := sz + !backup_size)
        (take (List.length stack) backup)
        stack;
      let body_instrs, body_sz =
        compile_expression [] false Boxed (ofs + !backup_size) body in
      let store_instrs, store_sz = astore idx in
      let handler_instrs, handler_sz =
        compile_expression [] false Boxed (ofs + !backup_size + body_sz + 3 + 1 + meth_size + store_sz) handler in
      exception_table := (ofs + !backup_size,
                          ofs + !backup_size + body_sz,
                          ofs + !backup_size + body_sz + 3)
        :: !exception_table;
      (List.concat !save_instrs)
      @ body_instrs
      @ [ Instruction.GOTO (check_jump (3 + 1 + meth_size + store_sz + handler_sz)) ]
      @ [ Instruction.ALOAD_0 ]
      @ meth_asValue
      @ store_instrs
      @ handler_instrs
      @ (if !backup_size <> 0 then
        gstore_instrs
        @ (List.concat (List.rev !restore_instrs))
        @ gload_instrs
      else
        [])
      @ (if boxed_result = Unboxed then meth_asLong else []),
      !backup_size + body_sz + 3 + 1 + meth_size + store_sz + handler_sz
        + (if !backup_size <> 0 then gstore_sz + !backup_size + gload_sz else 0)
        + (if boxed_result = Unboxed then meth_size else 0)
  | Mstaticfail (nfail, args) ->
      let args_instrs, args_sz =
        compile_expression_list stack Boxed ofs args in
      let dest = check_jump ((Hashtbl.find current_catches nfail) - (ofs + args_sz)) in
      args_instrs
      @ [ Instruction.GOTO dest ],
      args_sz + 3
  | Mstaticcatch (nfail, body, _, handler) ->
      Hashtbl.add current_catches nfail (ofs + 3);
      let handler_instrs, handler_sz =
        compile_expression stack false boxed_result (ofs + 3) handler in
      let body_instrs, body_sz =
        compile_expression stack false boxed_result (ofs + 3 + handler_sz + 3) body in
      [ Instruction.GOTO (check_jump (3 + handler_sz + 3)) ]
      @ handler_instrs
      @ [ Instruction.GOTO (check_jump (3 + body_sz)) ]
      @ body_instrs,
      3 + handler_sz + 3 + body_sz
  | Mswitch (arg, consts, actions, default) ->
      let consts_len = Array.length consts in
      let arg_instrs, arg_sz =
        compile_expression stack false boxed_result ofs arg in
      let switch_ofs = ofs + arg_sz in
      let padding = 3 - (switch_ofs mod 4) in
      let switch_sz = 1 + padding + 4 + 4 + 4 + (4 * consts_len) in
      let o = ref (ofs + arg_sz + switch_sz) in
      let actions_sz = ref 0 in
      let actions_len = Array.length actions in
      let compiled_actions = Array.make actions_len None in
      for i = 0 to pred actions_len do
        let act_instrs, act_sz =
          compile_expression stack is_tail boxed_result !o actions.(i) in
        compiled_actions.(i) <- Some (act_instrs, !o, !o + act_sz);
        actions_sz := !actions_sz + act_sz + 3;
        o := !o + act_sz + 3
      done;
      let default_instrs, default_sz =
        compile_expression stack false boxed_result !o default in
      o := !o + default_sz;
      arg_instrs
      @ [ Instruction.TABLESWITCH
            (Int32.of_int (!o - switch_ofs - default_sz),
             Int32.zero,
             Int32.of_int (pred consts_len),
             Array.to_list
               (Array.map
                  (fun idx ->
                    match compiled_actions.(idx) with
                    | Some (_, o, _) ->
                        Int32.of_int (o - switch_ofs)
                    | None -> Misc.fatal_error "Bytecodegen.compile_expression: error in switch compilation")
                  consts)) ]
      @ (List.concat
           (Array.to_list
              (Array.map
                 (function
                   | Some (instrs, _, ofs) ->
                       instrs
                       @ [ Instruction.GOTO (check_jump (!o - ofs)) ]
                   | None -> Misc.fatal_error "Bytecodegen.compile_expression: error in switch compilation")
                 compiled_actions)))
      @ default_instrs,
      arg_sz + switch_sz + !actions_sz + default_sz
  | Mint expr ->
      compile_expression stack false Unboxed ofs expr
  | Mtag expr ->
      let expr_instrs, expr_sz =
        compile_expression stack false Boxed ofs expr in
      expr_instrs
      @ meth_asBlock
      @ meth_getTag,
      expr_sz + meth_size + meth_size
and compile_expression_list stack boxed_results ofs l =
  let ofs', l', _ = List.fold_left
      (fun (o, l, ste) elem ->
        let res, sz =
          compile_expression ste false boxed_results o elem in
        (o + sz, res :: l, boxed_results :: ste))
      (ofs, [], stack)
      l in
  (List.concat (List.rev l'), ofs' - ofs)
and compile_expression_list_to_locals stack ofs l =
  let instrs = ref [] in
  let instrs_sz = ref 0 in
  let stores = ref [] in
  let stores_sz = ref 0 in
  let o = ref ofs in
  let st = ref stack in
  let idx = ref 1 in
  List.iter
    (fun elem ->
      let elem_instrs, elem_sz =
        compile_expression !st false Boxed !o elem in
      let store_instrs, store_sz = astore !idx in
      instrs := !instrs @ elem_instrs;
      instrs_sz := !instrs_sz + elem_sz;
      stores := store_instrs @ !stores;
      stores_sz := !stores_sz + store_sz;
      o := !o + elem_sz;
      st := Boxed :: !st;
      incr idx)
    l;
  !instrs @ !stores, !o - ofs + !stores_sz
and compile_block stack tag mut ofs args =
  let size = List.length args in
  if size = 0 && mut = Asttypes.Immutable then
    let tag_instrs, tag_sz = push_int size in
    [ Instruction.ALOAD_0 ]
    @ tag_instrs
    @ meth_getAtom,
    1 + tag_sz + meth_size
  else
    (match size with
    | 1 | 2 | 3 | 4 ->
        let args_instrs, args_sz =
          compile_expression_list stack Boxed ofs args in
        let tag_instrs, tag_sz = push_int tag in
        args_instrs
        @ tag_instrs
        @ [ Instruction.INVOKESTATIC
              (class_AbstractNativeRunner,
               make_method "createBlock",
               ((parameters size) @ [`Int], `Class class_Value)) ],
        args_sz + tag_sz + meth_size
    | _ ->
        let args_instrs, args_sz =
          compile_expression_list stack Boxed ofs args in
        let tag_instrs, tag_sz = push_int tag in
        let size_instrs, size_sz = push_int size in
        let fold_instrs = ref [] in
        let fold_sz = ref 0 in
        let remaining = ref size in
        while !remaining > 0 do
          let n = min 4 !remaining in
          fold_instrs := !fold_instrs
            @ [ Instruction.INVOKESTATIC
                  (class_AbstractNativeRunner,
                   make_method ("foldValues" ^ (string_of_int n)),
                   ((parameters n) @ [`Class class_Values],
                    `Class class_Values)) ];
          fold_sz := !fold_sz + meth_size;
          remaining := !remaining - n
        done;
        args_instrs
        @ tag_instrs
        @ size_instrs
        @ meth_createValues
        @ !fold_instrs
        @ meth_returnValues,
        args_sz + tag_sz + size_sz + meth_size + !fold_sz + meth_size)
and compile_first_closure cls func arity =
  let load_instrs, load_sz = aload !garbage_index in
  (if arity = 1 then
    load_instrs
    @ [ Instruction.DUP ;
        Instruction.ALOAD_0 ;
        Instruction.LDC_W (`Class_or_interface (make_class cls)) ;
        Instruction.LDC_W (`String (normalize_function_name func)) ]
    @ meth_registerClosure
    @ meth_setCode
    @ [ Instruction.ICONST_1 ]
    @ field_ONE
    @ meth_set,
    load_sz + 1 + 1 + 3 + 3 + meth_size + meth_size + 1 + field_size + meth_size,
    2
  else
    let lbl = if arity < 0 then -1 else -arity - 1 in
    let lbl_instrs, lbl_sz = push_int lbl in
    let arity_instrs, arity_sz = push_boxed_int arity in
    load_instrs
    @ [ Instruction.DUP ;
        Instruction.DUP ]
    @ lbl_instrs
    @ meth_setCode
    @ [ Instruction.ICONST_1 ]
    @ arity_instrs
    @ meth_set
    @ [ Instruction.ICONST_2 ;
        Instruction.ALOAD_0 ;
        Instruction.LDC_W (`Class_or_interface (make_class cls)) ;
        Instruction.LDC_W (`String (normalize_function_name func)) ]
    @ meth_registerClosure
    @ meth_createFromLong
    @ meth_set,
    load_sz + 1 + 1 + lbl_sz + meth_size + 1 + arity_sz + meth_size + 1 + 1
      + 3 + 3 + meth_size + meth_size + meth_size,
  3)
and compile_other_closures next = function
  | [] -> [], 0, next
  | others ->
      let load_instrs, load_sz = aload !garbage_index in
      let idx = ref next in
      let instrs = ref [] in
      let sz = ref 0 in
      List.iter
        (fun ((cls, func), arity) ->
          let idx_instrs, idx_sz = push_int !idx in
          if arity = 1 then begin
            instrs := !instrs
              @ [ Instruction.ALOAD_0 ]
              @ load_instrs
              @ idx_instrs
              @ [ Instruction.LDC_W (`Class_or_interface (make_class cls)) ;
                  Instruction.LDC_W (`String (normalize_function_name func)) ]
              @ meth_registerInfix1;
            sz := !sz + 1 + load_sz + idx_sz + 3 + 3 + meth_size;
            idx := !idx + 3
          end else begin
            let arity_instrs, arity_sz = push_int arity in
            instrs := !instrs
              @ [ Instruction.ALOAD_0 ]
              @ load_instrs
              @ idx_instrs
              @ [ Instruction.LDC_W (`Class_or_interface (make_class cls)) ;
                  Instruction.LDC_W (`String (normalize_function_name func)) ]
              @ arity_instrs
              @ meth_registerInfixN;
            sz := !sz + 1 + load_sz + idx_sz + 3 + 3 + arity_sz + meth_size;
            idx := !idx + 4
          end)
        others;
      !instrs, !sz, !idx

let compile_class meths =
  let classname = !Jclflags.java_package
    ^ "." ^ (Jcompilenv.current_unit_name ()) in
  let source_file =
    String.uncapitalize (Jcompilenv.current_unit_name ()) ^ ".ml" in
  let cd = { ClassDefinition.access_flags = [`Public; `Final; `Super];
             name = make_class classname;
             extends = Some class_Object;
             implements = [];
             fields = [];
             methods = meths;
             attributes = [`SourceFile (utf8_of_string source_file)]; } in
  let buff = Buffer.create 16 in
  let os = OutputStream.make_of_buffer buff in
  (try
    ClassFile.write (ClassDefinition.encode cd) os;
  with (BaristaLibrary.Attribute.Exception BaristaLibrary.Attribute.Invalid_code_length) ->
    raise (Error (Method_too_long classname)));
  OutputStream.close os;
  Buffer.contents buff

let compile_function fd =
  let rec signature acc = function
    | 0 -> (`Class class_AbstractNativeRunner) :: acc
    | n -> signature ((`Class class_Value) :: acc) (pred n) in
  current_file :=
    String.uncapitalize (Jcompilenv.current_unit_name ()) ^ ".ml";
  current_class := !Jclflags.java_package
    ^ "." ^ (Jcompilenv.current_unit_name ());
  current_function := fd.fun_name;
  exception_table := [];
  garbage_index := fd.fun_garbage_index;
  Hashtbl.clear current_catches;
  line_number_table := [];
  prev_ofs := 0;
  let parameters = signature [] (List.length fd.fun_args) in
  let instrs, _ = compile_expression [] true Boxed 0 fd.fun_body in
  let instrs = instrs @ [ Instruction.ARETURN ] in
  let exception_table =
    List.rev_map
      (fun (start_pc, end_pc, handler_pc) ->
        (start_pc, end_pc, handler_pc, (Some class_FailException)))
      !exception_table in
  let graph = ControlFlow.graph_of_instructions instrs exception_table in
  let graph =
    ControlFlow.map_graph
      (fun ofs instrs ->
        let ofs = Int32.to_int ofs in
        let sz = Instruction.size_of_list ofs instrs in
        let line_numbers = List.filter (fun (o, _) -> o >= ofs && o < ofs + sz) !line_number_table in
        let line_numbers = List.map (fun (o, x) -> o - ofs, x) line_numbers in
        line_numbers, instrs)
      (fun x _ -> x)
      (fun x _ _ -> x)
      graph in
  let graph = Code.remove_dead_code graph in
  let instrs, exception_table, graph = Code.flatten_graph graph in
  let line_numbers =
    ControlFlow.fold_vertices
      (fun v acc ->
        let ofs, l = ControlFlow.label_of_vertex v in
        let ofs = Int32.to_int ofs in
        let l = List.map (fun (o, x) -> o + ofs, x) l in
        acc @ l)
      graph
      [] in
  let graph = ControlFlow.map_graph (fun (x, _) z -> x, z) (fun x _ -> x) (fun x _ _ -> x) graph in
  let size = Instruction.size_of_list 0 instrs in
  let init_state = StackState.make_of_parameters None parameters in
  let unifier = StackState.unify_to_java_lang_Object in
  let max_stack, _, stack_map_frame =
    try
      Code.compute_stack_infos unifier graph init_state
    with
    | StackState.Exception err ->
        raise (Error (Unable_to_compute_stack (StackState.string_of_error err)))
    | e -> raise (Error (Unable_to_compute_stack (Printexc.to_string e))) in
  let code = { Attribute.max_stack = max_stack;
               Attribute.max_locals = fd.fun_max_locals;
               Attribute.code = instrs;
               Attribute.exception_table = exception_table;
               Attribute.attributes =
               (if (stack_map_frame <> []) && not (!Jclflags.no_stackmap) then
                 [ `StackMapTable stack_map_frame ]
               else
                 [])
               @ (if line_numbers <> [] then
                 [ `LineNumberTable (List.sort compare line_numbers) ]
               else
                 [])
               @ (if !Clflags.debug then
                 [ `LocalVariableTable
                     ((0,
                       size,
                       utf8_of_string "runner",
                       `Class class_AbstractNativeRunner,
                       0)
                      :: (List.rev
                            (snd (List.fold_left
                                    (fun (idx, l) id ->
                                      succ idx,
                                      ((0,
                                        size,
                                        (utf8_of_string (Ident.name id)),
                                        (`Class class_Value),
                                        idx) :: l))
                                    (1, [])
                                    fd.fun_args)))) ]
               else
                 []) ; } in
  Method.Regular ([`Public; `Static],
                  make_method fd.fun_name,
                  (parameters, `Class class_Value),
                  [`Code code;
                   `Exceptions [class_FailException; class_FatalException]])

let report_error ppf = function
  | Method_too_long name ->
      Format.fprintf ppf "Cannot compile %s (Java method is too long)" name
  | Unable_to_compute_stack msg ->
      Format.fprintf ppf "Cannot compute stack frames (%s)" msg
