(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://qt.nokia.com/doc/4.0/qpl.html>.
 *)

open BaristaLibrary
open Utils


(* Name functions *)

let normalize_function_name s =
  if (s = "<init>") || (s = "<clinit>") then
    utf8_of_string s
  else
    List.fold_left
      (fun acc (o, n) ->
        utf8_replace o n acc)
      (utf8_of_string s)
      [ (uchar_of_char '.', uchar_of_code 0x00B7) ;
        (uchar_of_char ';', uchar_of_code 0x00A6) ;
        (uchar_of_char '[', uchar_of_code 0x00A1) ;
        (uchar_of_char '/', uchar_of_code 0x00F7) ;
        (uchar_of_char '<', uchar_of_code 0x00AB) ;
        (uchar_of_char '>', uchar_of_code 0x00BB) ]

let make_class s =
  Name.make_for_class_from_external (utf8_of_string s)

let make_field s =
  Name.make_for_field (utf8_of_string s)

let make_method s =
  Name.make_for_method (normalize_function_name s)


(* Class constants *)

let class_AbstractNativeApplet = make_class "fr.x9c.cadmium.support.AbstractNativeApplet"
let class_AbstractNativeRunner = make_class "fr.x9c.cadmium.kernel.AbstractNativeRunner"
let class_AbstractNativeServlet = make_class "fr.x9c.cadmium.support.servlet.AbstractNativeServlet"
let class_Application = make_class "fr.x9c.cadmium.kernel.AbstractNativeRunner$Application"
let class_Block = make_class "fr.x9c.cadmium.kernel.Block"
let class_Class = make_class "java.lang.Class"
let class_Closure = make_class "fr.x9c.cadmium.kernel.AbstractNativeRunner$Closure"
let class_CodeRunner = make_class "fr.x9c.cadmium.kernel.CodeRunner"
let class_Fail = make_class "fr.x9c.cadmium.kernel.Fail"
let class_FailException = make_class "fr.x9c.cadmium.kernel.Fail$Exception"
let class_FatalException = make_class "fr.x9c.cadmium.kernel.Fatal$Exception"
let class_InputStream = make_class "java.io.InputStream"
let class_Map = make_class "java.util.Map"
let class_NativeParameters = make_class "fr.x9c.cadmium.kernel.NativeParameters"
let class_Object = make_class "java.lang.Object"
let class_PrintStream = make_class "java.io.PrintStream"
let class_String = make_class "java.lang.String"
let class_System = make_class "java.lang.System"
let class_Value = make_class "fr.x9c.cadmium.kernel.Value"
let class_Values = make_class "fr.x9c.cadmium.kernel.AbstractNativeRunner$Values"


(* Field constants *)

let field_DOUBLE_ARRAY_TAG = [ Instruction.GETSTATIC (class_Block, make_field "DOUBLE_ARRAY_TAG", `Int) ]
let field_DOUBLE_SIZE = [ Instruction.GETSTATIC (class_Block, make_field "DOUBLE_SIZE", `Int) ]
let field_ERR = [ Instruction.GETSTATIC (class_System, make_field "err", `Class class_PrintStream) ]
let field_FALSE = [ Instruction.GETSTATIC (class_Value, make_field "FALSE", `Class class_Value) ]
let field_IN = [ Instruction.GETSTATIC (class_System, make_field "in", `Class class_InputStream) ]
let field_MINUS_ONE = [ Instruction.GETSTATIC (class_Value, make_field "MINUS_ONE", `Class class_Value) ]
let field_MINUS_TWO = [ Instruction.GETSTATIC (class_Value, make_field "MINUS_TWO", `Class class_Value) ]
let field_ONE = [ Instruction.GETSTATIC (class_Value, make_field "ONE", `Class class_Value) ]
let field_OUT = [ Instruction.GETSTATIC (class_System, make_field "out", `Class class_PrintStream) ]
let field_TRUE = [ Instruction.GETSTATIC (class_Value, make_field "TRUE", `Class class_Value) ]
let field_TWO = [ Instruction.GETSTATIC (class_Value, make_field "TWO", `Class class_Value) ]
let field_UNIT = [ Instruction.GETSTATIC (class_Value, make_field "UNIT", `Class class_Value) ]
let field_ZERO = [ Instruction.GETSTATIC (class_Value, make_field "ZERO", `Class class_Value) ]
let field_applet_runner = [ Instruction.PUTFIELD (class_AbstractNativeApplet, make_field "runner", `Class class_AbstractNativeRunner) ]
let field_servlet_runner = [ Instruction.PUTSTATIC (class_AbstractNativeServlet, make_field "runner", `Class class_AbstractNativeRunner) ]

let field_size = 3
let field_stack = 1


(* Method constants *)

let cstr_AbstractNativeApplet =
  [ Instruction.INVOKESPECIAL
      (class_AbstractNativeApplet,
       make_method "<init>",
       ([], `Void)) ]
let cstr_AbstractNativeRunner =
  [ Instruction.INVOKESPECIAL
      (class_AbstractNativeRunner,
       make_method "<init>",
       ([`Class class_NativeParameters], `Void)) ]
let cstr_AbstractNativeServlet =
  [ Instruction.INVOKESPECIAL
      (class_AbstractNativeServlet,
       make_method "<init>",
       ([], `Void)) ]
let cstr_copy_AbstractNativeRunner =
  [ Instruction.INVOKESPECIAL
      (class_AbstractNativeRunner,
       make_method "<init>",
       ([`Class class_AbstractNativeRunner], `Void))]
let meth_asBlock =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Value),
       make_method "asBlock",
       ([], `Class class_Block)) ]
let meth_asCodeOffset =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Value),
       make_method "asCodeOffset",
       ([], `Int)) ]
let meth_asDouble =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Block),
       make_method "asDouble",
       ([], `Double)) ]
let meth_asLong =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Value),
       make_method "asLong",
       ([], `Int)) ]
let meth_asValue =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_FailException),
       make_method "asValue",
       ([`Class class_AbstractNativeRunner], `Class class_Value)) ]
let meth_caml_make_array =
  [ Instruction.INVOKESTATIC
      (make_class "fr.x9c.cadmium.primitives.stdlib.Array",
       make_method "caml_make_array",
       ([`Class class_CodeRunner; `Class class_Value], `Class class_Value)) ]
let meth_checkSignals =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_AbstractNativeRunner),
       make_method "checkSignals",
       ([], `Void)) ]
let meth_createApplication =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_AbstractNativeRunner),
       make_method "createApplication",
       ([`Class class_Value; `Int], `Class class_Application)) ]
let meth_createBlock =
  [ Instruction.INVOKESTATIC
      (class_Block,
       make_method "createBlock",
       ([`Int; `Int], (`Class class_Block))) ]
let meth_createClosure =
  [ Instruction.INVOKESTATIC
      (class_Block,
       make_method "createClosure",
       ([`Int], `Class class_Block)) ]
let meth_createClosureVars =
  [ Instruction.INVOKESTATIC
      (class_AbstractNativeRunner,
       make_method "createClosureVars",
       ([`Int; `Int], `Class class_Closure)) ]
let meth_createDouble =
  [ Instruction.INVOKESTATIC
      (class_Block,
       make_method "createDouble",
       ([`Double], `Class class_Block)) ]
let meth_createDoubles =
  [ Instruction.INVOKESTATIC
      (class_AbstractNativeRunner,
       make_method "createDoubles",
       ([`Int], `Class class_Values)) ]
let meth_createDoubleArray =
  [ Instruction.INVOKESTATIC
      (class_Block,
       make_method "createDoubleArray",
       ([`Int], `Class class_Block)) ]
let meth_createFromBlock =
  [ Instruction.INVOKESTATIC
      (class_Value,
       make_method "createFromBlock",
       ([`Class class_Block], `Class class_Value)) ]
let meth_createFromLong =
  [ Instruction.INVOKESTATIC
      (class_Value,
       make_method "createFromLong",
       ([`Int], `Class class_Value)) ]
let meth_createFromRawValue =
  [ Instruction.INVOKESTATIC
      (class_Value,
       make_method "createFromRawValue",
       ([`Int], `Class class_Value)) ]
let meth_createGlobal =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_AbstractNativeRunner),
       make_method "createGlobal",
       ([`Class class_String; `Int], `Void)) ]
let meth_createInfix =
  [ Instruction.INVOKESTATIC
      (class_Block,
       make_method "createInfix",
       ([`Int], `Class class_Block)) ]
let meth_createValues =
  [ Instruction.INVOKESTATIC
      (class_AbstractNativeRunner,
       make_method "createValues",
       ([`Int; `Int], `Class class_Values)) ]
let meth_equalValues =
  [ Instruction.INVOKESTATIC
      (class_AbstractNativeRunner,
       make_method "equalValues",
       ([`Class class_Value; `Class class_Value], `Boolean)) ]
let meth_execute =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_AbstractNativeRunner),
       make_method "execute",
       ([], `Void)) ]
let meth_executeWithBindings =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_AbstractNativeRunner),
       make_method "executeWithBindings",
       ([`Class class_Map], `Void)) ]
let meth_fromStream =
  [ Instruction.INVOKESTATIC
      (class_NativeParameters,
       make_method "fromStream",
       ([`Class class_InputStream ;
         `Array (`Class class_String) ;
         `Class class_InputStream ;
         `Class class_PrintStream ;
         `Class class_PrintStream ;
         `Boolean ;
         `Class (class_String)],
        `Class class_NativeParameters))]
let meth_get =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Block),
       make_method "get",
       ([`Int], `Class class_Value)) ]
let meth_getAtom =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_AbstractNativeRunner),
       make_method "getAtom",
       ([`Int], `Class class_Value)) ]
let meth_getConstant =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_AbstractNativeRunner),
       make_method "getConstant",
       ([`Class class_Class; `Int], `Class class_Value)) ]
let meth_getDouble =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Block),
       make_method "getDouble",
       ([`Int], `Double)) ]
let meth_getGlobal =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_AbstractNativeRunner),
       make_method "getGlobal",
       ([`Class class_String], (`Class class_Value))) ]
let meth_getHeader =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Block),
       make_method "getHeader",
       ([], `Int)) ]
let meth_getMagicNumber =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Block),
       make_method "getMagicNumber",
       ([], `Long)) ]
let meth_getRawValue =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Value),
       make_method "getRawValue",
       ([], `Int)) ]
let meth_getResourceAsStream =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Class),
       make_method "getResourceAsStream",
       ([`Class class_String], `Class class_InputStream)) ]
let meth_getResult =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_AbstractNativeRunner),
       make_method "getResult",
       ([], `Class class_Value)) ]
let meth_getTag =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Block),
       make_method "getTag",
       ([], `Int)) ]
let meth_getUnsignedByte =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Block),
       make_method "getUnsignedByte",
       ([`Int], `Int)) ]
let meth_getWoSize =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Block),
       make_method "getWoSize",
       ([], `Int)) ]
let meth_greaterEqualValue =
  [ Instruction.INVOKESTATIC
      (class_AbstractNativeRunner,
       make_method "greaterEqualValue",
       ([`Class class_Value; `Class class_Value], `Boolean)) ]
let meth_greaterThanValue =
  [ Instruction.INVOKESTATIC
      (class_AbstractNativeRunner,
       make_method "greaterThanValue",
       ([`Class class_Value; `Class class_Value], `Boolean)) ]
let meth_isBlock =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Value),
       make_method "isBlock",
       ([], `Boolean)) ]
let meth_isLong =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Value),
       make_method "isLong",
       ([], `Boolean)) ]
let meth_isOutValue =
  [ Instruction.INVOKESTATIC
      (class_AbstractNativeRunner,
       make_method "isOutValue",
       ([`Class class_Value; `Class class_Value], `Boolean)) ]
let meth_loadConstant =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_AbstractNativeRunner),
       make_method "loadConstant",
       ([`Class class_Class], `Void)) ]
let meth_lowerEqualValue =
  [ Instruction.INVOKESTATIC
      (class_AbstractNativeRunner,
       make_method "lowerEqualValue",
       ([`Class class_Value; `Class class_Value], `Boolean)) ]
let meth_lowerThanValue =
  [ Instruction.INVOKESTATIC
      (class_AbstractNativeRunner,
       make_method "lowerThanValue",
       ([`Class class_Value; `Class class_Value], `Boolean)) ]
let meth_notEqualValues =
  [ Instruction.INVOKESTATIC
      (class_AbstractNativeRunner,
       make_method "notEqualValues",
       ([`Class class_Value; `Class class_Value], `Boolean)) ]
let meth_offset =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Block),
       make_method "offset",
       ([`Int], `Class class_Value)) ]
let meth_raise =
  [ Instruction.INVOKESTATIC
      (class_Fail,
       make_method "raise",
       ([`Class class_Value], `Void)) ]
let meth_raiseZeroDivide =
  [ Instruction.INVOKESTATIC
      (class_Fail,
       make_method "raiseZeroDivide",
       ([], `Void)) ]
let meth_registerClosure =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_AbstractNativeRunner),
       make_method "registerClosure",
       ([`Class class_Class; `Class class_String], `Int)) ]
let meth_registerClosure1 =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_AbstractNativeRunner),
       make_method "registerClosure1",
       ([`Class class_Class; `Class class_String], `Class class_Value)) ]
let meth_registerClosureN =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_AbstractNativeRunner),
       make_method "registerClosureN",
       ([`Class class_Class; `Class class_String; `Int], `Class class_Value)) ]
let meth_registerInfix1 =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_AbstractNativeRunner),
       make_method "registerInfix1",
       ([`Class class_Block; `Int; `Class class_Class; `Class class_String],
        `Void)) ]
let meth_registerInfixN =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_AbstractNativeRunner),
       make_method "registerInfixN",
       ([`Class class_Block; `Int; `Class class_Class; `Class class_String; `Int],
        `Void)) ]
let meth_registerPredefinedException =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_AbstractNativeRunner),
       make_method "registerPredefinedException",
       ([`Class class_String], `Void)) ]
let meth_returnApplication =
  [ Instruction.INVOKESTATIC
      (class_AbstractNativeRunner,
       make_method "returnApplication",
       ([`Class class_Application], `Class class_Value)) ]
let meth_returnClosure =
  [ Instruction.INVOKESTATIC
      (class_AbstractNativeRunner,
       make_method "returnClosure",
       ([`Class class_Block; `Class class_Closure], `Class class_Value)) ]
let meth_returnDoubles =
  [ Instruction.INVOKESTATIC
      (class_AbstractNativeRunner,
       make_method "returnDoubles",
       ([`Class class_Values], `Class class_Value)) ]
let meth_returnValues =
  [ Instruction.INVOKESTATIC
      (class_AbstractNativeRunner,
       make_method "returnValues",
       ([`Class class_Values], `Class class_Value)) ]
let meth_set =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Block),
       make_method "set",
       ([`Int; `Class class_Value], `Void)) ]
let meth_setCode =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Block),
       make_method "setCode",
       ([`Int], `Void)) ]
let meth_setCustom =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Block),
       make_method "setCustom",
       ([`Class class_Object], `Void)) ]
let meth_setDouble =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Block),
       make_method "setDouble",
       ([`Int; `Double], `Void)) ]
let meth_setGlobal =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_AbstractNativeRunner),
       make_method "setGlobal",
       ([`Class class_String; `Class class_Value], `Void)) ]
let meth_setParent =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Block),
       make_method "setParent",
       ([`Class class_Block], `Void)) ]
let meth_setUnsignedByte =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Block),
       make_method "setUnsignedByte",
       ([`Int; `Int], `Void)) ]
let meth_signedToUnsigned =
  [ Instruction.INVOKESTATIC
      (make_class "fr.x9c.cadmium.util.Misc",
       make_method "signedToUnsigned",
       ([`Int], `Long)) ]
let meth_universalLong =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Value),
       make_method "universalLong",
       ([], `Long)) ]
let meth_universalUnsignedLong =
  [ Instruction.INVOKEVIRTUAL
      ((`Class_or_interface class_Value),
       make_method "universalUnsignedLong",
       ([], `Long)) ]

let meth_size = 3
let meth_stack m =
  let sz = function
    | `Array _
    | `Boolean
    | `Byte
    | `Char
    | `Class _
    | `Float
    | `Int
    | `Short -> 1
    | `Double
    | `Long -> 2
    | `Void -> 0 in
  let inst, (prm, ret) = match m with
  | [ Instruction.INVOKEDYNAMIC (_, d) ]
  | [ Instruction.INVOKEINTERFACE ((_, _, d) , _) ]
  | [ Instruction.INVOKESPECIAL (_, _, d) ]
  | [ Instruction.INVOKEVIRTUAL (_, _, d) ] -> 1, d
  | [ Instruction.INVOKESTATIC (_, _, d) ] -> 0, d
  | _ ->
      Misc.fatal_error "Bytecodeutils.meth_stack: invalid instruction list" in
  max
    0
    ((sz ret) - inst - (List.fold_left (fun acc elem -> acc + sz elem) 0 prm))


(* Utility functions *)

let push_int = function
  | -1 -> [ Instruction.ICONST_M1 ], 1
  | 0 -> [ Instruction.ICONST_0 ], 1
  | 1 -> [ Instruction.ICONST_1 ], 1
  | 2 -> [ Instruction.ICONST_2 ], 1
  | 3 -> [ Instruction.ICONST_3 ], 1
  | 4 -> [ Instruction.ICONST_4 ], 1
  | 5 -> [ Instruction.ICONST_5 ], 1
  | n when n >= -128 && n <= 127 -> [ Instruction.BIPUSH n ], 2
  | n when n >= -32768 && n <= 32767 -> [ Instruction.SIPUSH n ], 3
  | n -> [ Instruction.LDC_W (`Int (Int32.of_int n)) ], 3

let push_boxed_int = function
  | -2 -> field_MINUS_TWO, field_size
  | -1 -> field_MINUS_ONE, field_size
  | 0 -> field_ZERO, field_size
  | 1 -> field_ONE, field_size
  | 2 -> field_TWO, field_size
  | n ->
      let instrs, sz = push_int n in
      instrs
      @ meth_createFromLong,
      sz + meth_size

let aload = function
  | 0 -> [ Instruction.ALOAD_0 ], 1
  | 1 -> [ Instruction.ALOAD_1 ], 1
  | 2 -> [ Instruction.ALOAD_2 ], 1
  | 3 -> [ Instruction.ALOAD_3 ], 1
  | n when n >= 0 && n <= 255 -> [ Instruction.ALOAD n ], 2
  | n when n >= 0 && n <= 65535 -> [ Instruction.WIDE_ALOAD n ], 4
  | _ -> Misc.fatal_error "BytecodeUtils.aload: invalid index"

let astore = function
  | 0 -> [ Instruction.ASTORE_0 ], 1
  | 1 -> [ Instruction.ASTORE_1 ], 1
  | 2 -> [ Instruction.ASTORE_2 ], 1
  | 3 -> [ Instruction.ASTORE_3 ], 1
  | n when n >= 0 && n <= 255 -> [ Instruction.ASTORE n ], 2
  | n when n >= 0 && n <= 65535 -> [ Instruction.WIDE_ASTORE n ], 4
  | _ -> Misc.fatal_error "BytecodeUtils.astore: invalid index"

let iload = function
  | 0 -> [ Instruction.ILOAD_0 ], 1
  | 1 -> [ Instruction.ILOAD_1 ], 1
  | 2 -> [ Instruction.ILOAD_2 ], 1
  | 3 -> [ Instruction.ILOAD_3 ], 1
  | n when n >= 0 && n <= 255 -> [ Instruction.ILOAD n ], 2
  | n when n >= 0 && n <= 65535 -> [ Instruction.WIDE_ILOAD n ], 4
  | _ -> Misc.fatal_error "BytecodeUtils.iload: invalid index"

let istore = function
  | 0 -> [ Instruction.ISTORE_0 ], 1
  | 1 -> [ Instruction.ISTORE_1 ], 1
  | 2 -> [ Instruction.ISTORE_2 ], 1
  | 3 -> [ Instruction.ISTORE_3 ], 1
  | n when n >= 0 && n <= 255 -> [ Instruction.ISTORE n ], 2
  | n when n >= 0 && n <= 65535 -> [ Instruction.WIDE_ISTORE n ], 4
  | _ -> Misc.fatal_error "BytecodeUtils.istore: invalid index"

let iinc idx incr =
  match idx, incr with
  | n, n' when n >= 0 && n <= 255 && n' >= -128 && n' <= 127 ->
      [ Instruction.IINC (n, n') ], 3
  | n, n' when n >= 0 && n <= 65535 && n' >= -32768 && n' <= 32767 ->
      [ Instruction.WIDE_IINC (n, n') ], 6
  | _ -> Misc.fatal_error "BytecodeUtils.iinc: invalid index or increment"
