(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2010 Xavier Clerc.
 * Original file (asmcomp/cmmgen.ml in the Objective Caml source
 * distribution) is Copyright (C) INRIA.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://qt.nokia.com/doc/4.0/qpl.html>.
 *)

open Asttypes
open Lambda
open Jlambda
open Macroinstr


(* method frames *)

let max_locals = ref 2

type frame =
    { elements : (boxed * int) Ident.tbl;
      garbage : int;
      next : int; }

let make_frame vars =
  let nxt = ref 1 in (* slot 0 is for code runner *)
  let elems = List.fold_left
      (fun acc elem ->
        let idx = !nxt in
        incr nxt;
        Ident.add elem (Boxed, idx) acc)
      Ident.empty
      vars in
  max_locals := 1 + !nxt;
  { elements = elems;
    garbage = !nxt;
    next = 1 + !nxt; }

let reset_frame () =
  max_locals := 2

let add_var boxed id frm =
  let res = { elements = Ident.add id (boxed, frm.next) frm.elements;
              garbage = frm.garbage;
              next = succ frm.next; } in
  max_locals := max !max_locals res.next;
  frm.next, res


(* primitives *)

let default_prim name =
  { Primitive.prim_name = name;
    prim_arity = 0;
    prim_alloc = true;
    prim_native_name = "";
    prim_native_float =  false; }

let simplif_primitive = function
  | Pduprecord _ ->
      Pccall (default_prim "caml_obj_dup")
  | Pbigarrayref (_, n, Pbigarray_unknown, _) ->
      Pccall (default_prim ("caml_ba_get_" ^ string_of_int n))
  | Pbigarrayset (_ , n, Pbigarray_unknown, _) ->
      Pccall (default_prim ("caml_ba_set_" ^ string_of_int n))
  | Pbigarrayref (_, n, _, Pbigarray_unknown_layout) ->
      Pccall (default_prim ("caml_ba_get_" ^ string_of_int n))
  | Pbigarrayset (_, n, _, Pbigarray_unknown_layout) ->
      Pccall (default_prim ("caml_ba_set_" ^ string_of_int n))
  | Pbintofint Pint64 -> Pccall (default_prim "caml_int64_of_int")
  | Pintofbint Pint64 -> Pccall (default_prim "caml_int64_to_int")
  | Pcvtbint (Pint32, Pint64) -> Pccall (default_prim "caml_int64_of_int32")
  | Pcvtbint (Pint64, Pint32) -> Pccall (default_prim "caml_int64_to_int32")
  | Pcvtbint (Pnativeint, Pint64) -> Pccall (default_prim "caml_int64_of_nativeint")
  | Pcvtbint (Pint64, Pnativeint) -> Pccall (default_prim "caml_int64_to_nativeint")
  | Pnegbint Pint64 -> Pccall (default_prim "caml_int64_neg")
  | Paddbint Pint64 -> Pccall (default_prim "caml_int64_add")
  | Psubbint Pint64 -> Pccall (default_prim "caml_int64_sub")
  | Pmulbint Pint64 -> Pccall (default_prim "caml_int64_mul")
  | Pdivbint Pint64 -> Pccall (default_prim "caml_int64_div")
  | Pmodbint Pint64 -> Pccall (default_prim "caml_int64_mod")
  | Pandbint Pint64 -> Pccall (default_prim "caml_int64_and")
  | Porbint Pint64 -> Pccall (default_prim "caml_int64_or")
  | Pxorbint Pint64 -> Pccall (default_prim "caml_int64_xor")
  | Plslbint Pint64 -> Pccall (default_prim "caml_int64_shift_left")
  | Plsrbint Pint64 -> Pccall (default_prim "caml_int64_shift_right_unsigned")
  | Pasrbint Pint64 -> Pccall (default_prim "caml_int64_shift_right")
  | Pbintcomp (Pint64, Lambda.Ceq) -> Pccall (default_prim "caml_equal")
  | Pbintcomp (Pint64, Lambda.Cneq) -> Pccall (default_prim "caml_notequal")
  | Pbintcomp (Pint64, Lambda.Clt) -> Pccall (default_prim "caml_lessthan")
  | Pbintcomp (Pint64, Lambda.Cgt) -> Pccall (default_prim "caml_greaterthan")
  | Pbintcomp (Pint64, Lambda.Cle) -> Pccall (default_prim "caml_lessequal")
  | Pbintcomp (Pint64, Lambda.Cge) -> Pccall (default_prim "caml_greaterequal")
  | Pbigarrayref (_, n, Pbigarray_int64, _) -> Pccall (default_prim ("caml_ba_get_" ^ string_of_int n))
  | Pbigarrayset (_, n, Pbigarray_int64, _) -> Pccall (default_prim ("caml_ba_set_" ^ string_of_int n))
  | p -> p


(* recursive values *)

let fundecls_size fundecls =
  let sz = ref (-1) in
  List.iter
    (fun (_, arity, _, _) -> sz := !sz + 1 + (if arity = 1 then 2 else 3))
    fundecls;
  !sz

type rhs_kind =
  | RHS_block of int
  | RHS_nonrec

let rec expr_size = function
  | Jclosure (fundecls, clos_vars) ->
      RHS_block (fundecls_size fundecls + List.length clos_vars)
  | Jlet (_, _, body) ->
      expr_size body
  | Jletrec (_, body) ->
      expr_size body
  | Jprim (Pmakeblock _, args, _) ->
      RHS_block (List.length args)
  | Jprim (Pmakearray (Paddrarray | Pintarray), args, _) ->
      RHS_block (List.length args)
  | Jsequence (_, j) ->
      expr_size j
  | _ ->
      RHS_nonrec

(* translation from jlambda to macroinstr *)

let functions = (Queue.create () : (string * Ident.t list * jlambda) Queue.t)

let rec remove_unit = function
  | Munit -> Mnop
  | Msequence (m, Munit) -> m
  | Msequence (m1, m2) -> Msequence (m1, remove_unit m2)
  | Mifthenelse (cond, ifso, ifno) ->
      Mifthenelse (cond, remove_unit ifso, remove_unit ifno)
  | Mswitch (arg, index, cases, default) ->
      Mswitch (arg, index, Array.map remove_unit cases, remove_unit default)
  | Mstaticcatch (nfail, body, n, handler) ->
      Mstaticcatch (nfail, remove_unit body, n, remove_unit handler)
  | Mtrywith (body, exn, handler, backup) ->
      Mtrywith (remove_unit body, exn, remove_unit handler, backup)
  | m -> Msequence (m, Mpop)

let return_unit x =
  Msequence (x, Munit)

let current_class = ref ""

let rec imap n f = function
  | [] -> []
  | hd :: tl ->
      let hd' = f n hd in
      hd' :: (imap (succ n) f tl)

let rec transl stack frm = function
  | Jvar id ->
      let boxed, index = Ident.find_same id frm.elements in
      if boxed = Boxed then
        Mreadlocal index
      else
        Mreadlocalint index
  | Jconst c -> Mconst c
  | Jclosure (fundecls, clos_vars) ->
      List.iter
        (fun ((cls, func), _, params, body) ->
          if cls = !current_class then
            Queue.add (func, params, body) functions)
        fundecls;
      Mclosure (fundecls_size fundecls + List.length clos_vars,
                List.map
                  (fun (l, arity, _, _) ->
                    l, arity)
                  fundecls,
                imap 1 (fun n e -> (transl (stack + n) frm e)) clos_vars)
  | Joffset (j, n) ->
      let j' = transl stack frm j in
      if n = 0 then
        j'
      else
        Moffset (j', n)
  | Jdirect_apply ((cls, func), args, dbg) ->
        Mcall (Native,
               (cls, func),
               imap 1 (fun n e -> transl (stack + n) frm e) args,
               dbg)
  | Jgeneric_apply (clos, args, dbg) ->
      let arity = pred (List.length args) in
      let start = if arity <= 4 then 1 else 0 in
      let args' = imap start (fun n e -> transl (stack + n) frm e) (args @ [clos]) in
      Mapply (args', dbg)
  | Jsend (kind, meth, obj, args, dbg) ->
      let obj' = transl stack frm obj in
      let idx_obj, frm' = add_var Boxed (Ident.create "_dummy_") frm in
      (match kind, args with
      | Self, _ ->
          let arity = succ (List.length args) in
          let start = if arity <= 4 then 1 else 0 in
          let idx_meth, frm'' = add_var Boxed (Ident.create "_dummy_") frm' in
          let i1 = Mwritelocal (idx_obj, obj') in
          let i2 = Mwritelocal (idx_meth,
                                Mdynamicoffset ((Mreadlocal idx_obj),
                                                (transl stack frm' meth))) in
          let i3 = Mapply ((Mreadlocal idx_obj)
                           :: (imap (start + 1) (fun n e -> transl (stack + n) frm'' e) args)
                           @ [Mreadlocal idx_meth],
                           dbg) in
          Msequence (i1, Msequence (i2, i3))
      | Cached, _ :: _ :: args
      | _, args ->
          let arity = succ (List.length args) in
          let start = if arity <= 4 then 1 else 0 in
          let idx_meth, frm'' = add_var Boxed (Ident.create "_dummy_") frm' in
          let i1 = Mwritelocal (idx_obj, obj') in
          let i2 = Mwritelocal (idx_meth,
                                Mcall (Primitive,
                                       ("fr.x9c.cadmium.primitives.stdlib.Obj",
                                        "caml_get_public_method"),
                                       [Mreadlocal idx_obj;
                                        (transl stack frm' meth)],
                                       Debuginfo.none)) in
          let i3 = Mapply ((Mreadlocal idx_obj)
                           :: (imap (start + 1) (fun n e -> transl (stack + n) frm'' e) args)
                           @ [Mreadlocal idx_meth],
                           dbg) in
          Msequence (i1, Msequence (i2, i3)))
  | Jlet (id, expr, body) ->
      let expr' = transl stack frm expr in
      let idx, frm' = add_var Boxed id frm in
      Msequence (Mwritelocal (idx, expr'),
                 transl stack frm' body)
  | Jletrec (bindings, body) ->
      transl_letrec stack frm bindings body
  | Jprim (prim, args, dbg) ->
      let prim' = simplif_primitive prim in
      let is_call = Bytecodeprimitive.is_mcall prim' in
      Mprim (prim',
             imap
               (if is_call then 1 else 0)
               (fun n e -> transl (stack + n) frm e) args,
             dbg)
  | Jswitch (arg, sw) ->
      if Array.length sw.js_index_blocks = 0 then begin
        if Array.length sw.js_actions_consts = 1 then
          transl stack frm sw.js_actions_consts.(0)
        else
          Mswitch (Mint (transl stack frm arg),
                   sw.js_index_consts,
                   Array.map (transl stack frm) sw.js_actions_consts,
                   Munit)
      end else if Array.length sw.js_index_consts = 0 then begin
        if Array.length sw.js_actions_blocks = 1 then
          transl stack frm sw.js_actions_blocks.(0)
        else
          Mswitch (Mtag (transl stack frm arg),
                   sw.js_index_blocks,
                   Array.map (transl stack frm) sw.js_actions_blocks,
                   Munit)
      end else begin
        let idx, frm' = add_var Boxed (Ident.create "_dummy_") frm in
        let i1 = Mwritelocal (idx, transl stack frm arg) in
        let sw_consts = if Array.length sw.js_actions_consts = 1 then
          transl stack frm' sw.js_actions_consts.(0)
        else
          Mswitch (Mint (Mreadlocal idx),
                   sw.js_index_consts,
                   Array.map (transl stack frm') sw.js_actions_consts,
                   Munit) in
        let sw_blocks = if Array.length sw.js_actions_blocks = 1 then
          transl stack frm' sw.js_actions_blocks.(0)
        else
          Mswitch (Mtag (Mreadlocal idx),
                   sw.js_index_blocks,
                   Array.map (transl stack frm') sw.js_actions_blocks,
                   Munit) in
        let i2 = Mifthenelse (Mprim (Pisint, [Mreadlocal idx], Debuginfo.none),
                              sw_consts,
                              sw_blocks) in
        Msequence (i1, i2)
      end
  | Jstaticfail (nfail, args) ->
      Mstaticfail (nfail, imap 0 (fun n e -> transl (stack + n) frm e) args)
  | Jstaticcatch (nfail, [], body, handler) ->
      make_catch nfail (transl stack frm body) (transl stack frm handler)
  | Jstaticcatch (nfail, ids, body, handler) ->
      let frm', instrs = List.fold_left
          (fun (frm, m) elem ->
            let idx, frm' = add_var Boxed elem frm in
            (frm', Msequence (Mpoptolocal idx, m)))
          (frm, Mnop)
          ids in
      Mstaticcatch (nfail,
                    transl stack frm body,
                    (List.length ids),
                    Msequence (instrs, (transl stack frm' handler)))
  | Jtrywith (body, id, handler) ->
      let dummy = Ident.create "_dummy_" in
      let backup = ref [] in
      let new_frm = ref frm in
      for i = 1 to stack do
        let id, f = add_var Boxed dummy !new_frm in
        new_frm := f;
        backup := id :: !backup
      done;
      let idx, frm' = add_var Boxed id !new_frm in
      Mtrywith ((transl 0 !new_frm body),
                idx,
                (transl 0 frm' handler),
                !backup)
  | Jifthenelse (Jprim (Pnot, [arg], _), ifso, ifno) ->
      transl stack frm (Jifthenelse (arg, ifno, ifso))
  | Jifthenelse (cond, ifso, Jstaticfail (nfail, [])) ->
      exit_if_false stack frm cond (transl stack frm ifso) nfail
  | Jifthenelse (cond, Jstaticfail (nfail, []), ifno) ->
      exit_if_true stack frm cond nfail (transl stack frm ifno)
  | Jifthenelse (Jprim (Psequand, _, _) as cond, ifso, ifno) ->
      let raise_num = next_raise_count () in
      make_catch
        raise_num
        (exit_if_false stack frm cond (transl stack frm ifso) raise_num)
        (transl stack frm ifno)
  | Jifthenelse (Jprim (Psequor, _, _) as cond, ifso, ifno) ->
      let raise_num = next_raise_count () in
      make_catch
        raise_num
        (exit_if_true stack frm cond raise_num (transl stack frm ifno))
        (transl stack frm ifso)
  | Jifthenelse (Jifthenelse (cond, condso, condno), ifso, ifno) ->
      let num_true = next_raise_count () in
      make_catch
        num_true
        (make_catch2
           (fun shared_false ->
             Mifthenelse (transl stack frm cond,
                          exit_if_true stack frm condso num_true shared_false,
                          exit_if_true stack frm condno num_true shared_false))
           (transl stack frm ifno))
        (transl stack frm ifso)
  | Jifthenelse (cond, ifso, ifno) ->
      Mifthenelse (transl stack frm cond,
                   transl stack frm ifso,
                   transl stack frm ifno)
  | Jsequence (j1, j2) ->
      Msequence (remove_unit (transl stack frm j1), transl stack frm j2)
  | Jwhile (cond, body) ->
      return_unit (Mwhile (transl stack frm cond,
                           remove_unit (transl stack frm body)))
  | Jfor (id, low, high, dir, body) ->
      let low' = transl stack frm low in
      let idx, frm' = add_var Unboxed id frm in
      let high' = transl stack frm' high in
      let idx_high, frm'' = add_var Unboxed (Ident.create "_dummy_") frm' in
      let body' = remove_unit (transl stack frm'' body) in
      let loop =
        Msequence (Mwritelocalint (idx, low'),
                   (Msequence (Mwritelocalint (idx_high, high'),
                               (Mfor (idx, dir, idx_high, body'))))) in
      return_unit loop
  | Jassign (id, expr) ->
      let expr' = transl stack frm expr in
      let _, idx = Ident.find_same id frm.elements in
      return_unit (Mwritelocal (idx, expr'))
and make_catch nfail body handler =
  match body with
  | Mstaticfail (nfail', []) when nfail = nfail' -> handler
  | _ -> Mstaticcatch (nfail, body, 0, handler)
and make_catch2 mk_body handler =
  match handler with
  | Mstaticfail (_, [])
  | Mconst (Const_base (Const_int _))
  | Mconst (Const_base (Const_char _))
  | Mconst (Const_pointer _) ->
      mk_body handler
  | _ ->
      let nfail = next_raise_count () in
      make_catch
        nfail
        (mk_body (Mstaticfail (nfail, [])))
        handler
and exit_if_true stack frm cond nfail otherwise =
  match cond with
  | Jconst (Const_pointer 0) -> otherwise
  | Jconst (Const_pointer 1) -> Mstaticfail (nfail, [])
  | Jprim (Psequor, [arg1; arg2], _) ->
      exit_if_true stack frm arg1 nfail
        (exit_if_true stack frm arg2 nfail otherwise)
  | Jprim (Psequand, _, _) ->
      (match otherwise with
      | Mstaticfail (raise_num, []) ->
          exit_if_false stack frm cond (Mstaticfail (nfail, [])) raise_num
      | _ ->
          let raise_num = next_raise_count () in
          make_catch
            raise_num
            (exit_if_false stack frm cond (Mstaticfail (nfail, [])) raise_num)
            otherwise)
  | Jprim (Pnot, [arg], _) ->
      exit_if_false stack frm arg otherwise nfail
  | Jifthenelse (cond, ifso, ifno) ->
      make_catch2
        (fun shared ->
          Mifthenelse (transl stack frm cond,
                       exit_if_true stack frm ifso nfail shared,
                       exit_if_true stack frm ifno nfail shared))
        otherwise
  | _ ->
      Mifthenelse (transl stack frm cond, Mstaticfail (nfail, []), otherwise)
and exit_if_false stack frm cond otherwise nfail =
  match cond with
  | Jconst (Const_pointer 0) -> Mstaticfail (nfail, [])
  | Jconst (Const_pointer 1) -> otherwise
  | Jprim (Psequand, [arg1; arg2], _) ->
      exit_if_false stack frm arg1
        (exit_if_false stack frm arg2 otherwise nfail) nfail
  | Jprim (Psequor, _, _) ->
      (match otherwise with
      | Mstaticfail (raise_num, []) ->
          exit_if_true stack frm cond raise_num (Mstaticfail (nfail, []))
      | _ ->
          let raise_num = next_raise_count () in
          make_catch
            raise_num
            (exit_if_true stack frm cond raise_num (Mstaticfail (nfail, [])))
            otherwise)
  | Jprim (Pnot, [arg], _) ->
      exit_if_true stack frm arg nfail otherwise
  | Jifthenelse (cond, ifso, ifno) ->
      make_catch2
        (fun shared ->
          Mifthenelse (transl stack frm cond,
                       exit_if_false stack frm ifso shared nfail,
                       exit_if_false stack frm ifno shared nfail))
        otherwise
  | _ ->
      Mifthenelse (transl stack frm cond, otherwise, Mstaticfail (nfail, []))
and transl_letrec stack frame bindings body =
  let bsz = List.map (fun (id, exp) -> (id, exp, expr_size exp)) bindings in
  let rec init_blocks frm = function
    | [] ->
        fill_nonrec frm bsz
    | (id, _, RHS_block sz) :: tail ->
        let idx, frm' = add_var Boxed id frm in
        let expr = Mprim (Pccall (default_prim "caml_alloc_dummy"),
                                  [Mconst (Const_base (Const_int sz))],
                                  Debuginfo.none) in
        Msequence (Mwritelocal (idx, expr),
                   init_blocks frm' tail)
    | (id, _, RHS_nonrec) :: tail ->
        let idx, frm' = add_var Boxed id frm in
        Msequence (Mwritelocal (idx, Mconst (Const_base (Const_int 0))),
                   init_blocks frm' tail)
  and fill_nonrec frm = function
    | [] -> fill_blocks frm bsz
    | (_, _, RHS_block _) :: tail -> fill_nonrec frm tail
    | (id, exp, RHS_nonrec) :: tail ->
        let expr = transl stack frm exp in
        let idx, frm' = add_var Boxed id frm in
        Msequence (Mwritelocal (idx, expr),
                   fill_nonrec frm' tail)
  and fill_blocks frm = function
    | [] -> transl stack frm body
    | (id, exp, RHS_block _) :: tail ->
        let _, idx = Ident.find_same id frm.elements in
        let updt = Mprim (Pccall (default_prim "caml_update_dummy"),
                                  [Mreadlocal idx; transl stack frm exp],
                                  Debuginfo.none) in
        Msequence (Msequence (updt, Mpop), fill_blocks frm tail)
    | (_, _, RHS_nonrec) :: tail -> fill_blocks frm tail in
  init_blocks frame bsz

(* functions *)

let transl_function lbl params body =
  reset_frame ();
  let frm = make_frame params in
  let body' = transl 0 frm body in
  { fun_name = lbl;
    fun_args = params;
    fun_body = body';
    fun_max_locals = !max_locals;
    fun_garbage_index = frm.garbage; }

module StringSet = Set.Make (struct type t = string let compare = compare end)

let rec transl_all_functions already_translated cont =
  try
    let lbl, params, body = Queue.take functions in
    if StringSet.mem lbl already_translated then
      transl_all_functions already_translated cont
    else
      transl_all_functions
        (StringSet.add lbl already_translated)
        (transl_function lbl params body :: cont)
  with
  | Queue.Empty -> cont


(* whole unit *)

let compunit size jlam =
  let glob = Jcompilenv.make_symbol None in
  reset_frame ();
  let frm = make_frame [] in
  let name = Jcompilenv.make_symbol (Some "entry") in
  let init_code = transl 0 frm jlam in
  let entry = { fun_name = name;
                fun_args = [];
                fun_body = Msequence (Mcreateglobal (glob, size), init_code);
                fun_max_locals = !max_locals;
                fun_garbage_index = frm.garbage; } in
  transl_all_functions StringSet.empty [entry]

let main_return x =
  if !Jclflags.scripting then
    match x with
    | Jsequence (y, _) -> y
    | _ -> x
  else
    x

let compile_implementation prefixname ppf (size, lam) =
  let print_if flag printer arg =
    if !flag then Format.fprintf ppf "%a@." printer arg in
  current_class := !Jclflags.java_package
    ^ "." ^ (Jcompilenv.current_unit_name ());
  let class_path = String.copy !current_class in
  for i = 0 to pred (String.length class_path) do
    if class_path.[i] = '.' then class_path.[i] <- '/'
  done;
  let jlam = main_return (Jclosure.jlambda_of_lambda size lam) in
  print_if Jclflags.dump_jlambda Printjlambda.jlambda jlam;
  let functions = compunit size jlam in
  print_if Jclflags.dump_minstr Printminstr.minstr functions;
  Bytecodegen.reset_constants ();
  let methods = List.map Bytecodegen.compile_function functions in
  let class_data = Bytecodegen.compile_class methods in
  let consts_data = Bytecodegen.compile_constants () in
  let zip = Zip.open_out (prefixname ^ Jconfig.ext_obj) in
  Zip.add_entry class_data zip (class_path ^ ".class");
  Zip.add_entry consts_data zip (class_path ^ ".consts");
  Zip.close_out zip
