(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2010 Xavier Clerc.
 * Original file (asmcomp/asmlibrarian.ml in the Objective Caml source
 * distribution) is Copyright (C) INRIA.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://qt.nokia.com/doc/4.0/qpl.html>.
 *)

open Jcompilenv


type error =
  | File_not_found of string
  | Archiver_error of string

exception Error of error


let read_info name =
  let filename =
    try
      Misc.find_in_path !Config.load_path name
    with
    | Not_found -> raise (Error (File_not_found name)) in
  let info, crc = read_unit_info filename in
  info.ui_approx <- Jlambda.Value_unknown;
  (Filename.chop_suffix filename ".cmj" ^ Jconfig.ext_obj), (info, crc)

let create_archive file_names lib_name =
  let jar_name = Misc.chop_extension_if_any lib_name ^ Jconfig.ext_lib in
  let oc = open_out_bin lib_name in
  try
    output_string oc Jconfig.cmja_magic_number;
    let objfile_list, descr_list =
      List.split (List.map read_info file_names) in
    List.iter2
      (fun file_name (unit, crc) ->
        Javalink.check_consistency file_name unit crc)
      file_names
      descr_list;
    let infos = { lib_units = descr_list;
                  lib_ccobjs = !Clflags.ccobjs; } in
    output_value oc infos;
    (try
      let zip = Zip.open_out jar_name in
      List.iter
        (fun fn ->
          let z = Zip.open_in fn in
          List.iter
            (fun e ->
              if not e.Zip.is_directory then
                let data = Zip.read_entry z e in
                Zip.add_entry data zip e.Zip.filename)
            (Zip.entries z);
          Zip.close_in z)
        objfile_list;
      Zip.close_out zip
    with
    | _ -> raise (Error (Archiver_error jar_name)));
    close_out oc
  with x ->
    close_out oc;
    Misc.remove_file lib_name;
    Misc.remove_file jar_name;
    raise x

let report_error ppf = function
  | File_not_found name ->
      Format.fprintf ppf "Cannot find file %s" name
  | Archiver_error name ->
      Format.fprintf ppf "Error while creating the library %s" name
