(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2010 Xavier Clerc.
 * Original file (asmcomp/asmlink.ml in the Objective Caml source
 * distribution) is Copyright (C) INRIA.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://qt.nokia.com/doc/4.0/qpl.html>.
 *)

open Bytecodeutils
open Jcompilenv
open BaristaLibrary
open Utils


type error =
  | File_not_found of string
  | Not_an_object_file of string
  | Missing_implementations of (string * string list) list
  | Inconsistent_interface of string * string * string
  | Inconsistent_implementation of string * string * string
  | Linking_error
  | Multiple_definition of string * string * string
  | Missing_cmj of string * string
  | Unable_to_add_class of string
  | Unable_to_add_file of string
  | Duplicate_entry of string

exception Error of error


let crc_interfaces = Consistbl.create ()
let crc_implementations = Consistbl.create ()
let extra_implementations : string list ref = ref []
let implementations_defined : (string * string) list ref = ref []
let cmj_required : string list ref = ref []

let check_consistency file_name unit crc =
  (try
    List.iter
      (fun (name, crc) ->
        if name = unit.ui_name then
          Consistbl.set crc_interfaces name crc file_name
        else
          Consistbl.check crc_interfaces name crc file_name)
      unit.ui_imports_cmi
  with Consistbl.Inconsistency (name, user, auth) ->
    raise (Error (Inconsistent_interface (name, user, auth))));
  (try
    List.iter
      (fun (name, crc) ->
        if crc <> cmj_not_found_crc then
          Consistbl.check crc_implementations name crc file_name
        else if List.mem name !cmj_required then
          raise (Error (Missing_cmj (file_name, name)))
        else
          extra_implementations := name :: !extra_implementations)
      unit.ui_imports_cmj
  with Consistbl.Inconsistency (name, user, auth) ->
    raise (Error (Inconsistent_implementation (name, user, auth))));
  (try
    let source = List.assoc unit.ui_name !implementations_defined in
    raise (Error (Multiple_definition (unit.ui_name, file_name, source)))
  with Not_found ->
    ());
  Consistbl.set crc_implementations unit.ui_name crc file_name;
  implementations_defined :=
    (unit.ui_name, file_name) :: !implementations_defined;
  if unit.ui_symbol <> unit.ui_name then
    cmj_required := unit.ui_name :: !cmj_required

let extract_crc_interfaces () =
  Consistbl.extract crc_interfaces

let extract_crc_implementations () =
  List.fold_left
    (fun ncl n ->
      if List.mem_assoc n ncl then ncl else (n, cmj_not_found_crc) :: ncl)
    (Consistbl.extract crc_implementations)
    !extra_implementations

let missing_globals : (string, string list ref) Hashtbl.t = Hashtbl.create 17

let is_required name =
  try
    ignore (Hashtbl.find missing_globals name);
    true
  with
  | Not_found -> false

let add_required by (name, _) =
  try
    let rq = Hashtbl.find missing_globals name in
    rq := by :: !rq
  with
  | Not_found -> Hashtbl.add missing_globals name (ref [by])

let remove_required name =
  Hashtbl.remove missing_globals name

let extract_missing_globals () =
  let mg = ref [] in
  Hashtbl.iter (fun md rq -> mg := (md, !rq) :: !mg) missing_globals;
  !mg

type file =
  | Unit of string * Jcompilenv.unit_infos * Digest.t
  | Library of string * Jcompilenv.library_infos

let read_file obj_name =
  let file_name =
    try
      Misc.find_in_path !Config.load_path obj_name
    with Not_found -> raise (Error (File_not_found obj_name)) in
  if Filename.check_suffix file_name ".cmj" then begin
    let info, crc = Jcompilenv.read_unit_info file_name in
    Unit (file_name, info, crc)
  end else if Filename.check_suffix file_name ".cmja" then begin
    let infos =
      try
        Jcompilenv.read_library_info file_name
      with Jcompilenv.Error (Not_a_unit_info _) -> raise (Error (Not_an_object_file file_name)) in
    Library (file_name, infos)
  end else
    raise (Error (Not_an_object_file file_name))

let scan_file obj_name tolink =
  match read_file obj_name with
  | Unit (file_name, info, crc) ->
      remove_required info.ui_name;
      List.iter (add_required file_name) info.ui_imports_cmj;
      (info, file_name, crc) :: tolink
  | Library (file_name, infos) ->
      List.fold_right
        (fun (info, crc) acc ->
          remove_required info.ui_name;
          let file_name' = Printf.sprintf "%s(%s)" file_name info.ui_name in
          List.iter (add_required file_name') info.ui_imports_cmj;
          (info, file_name, crc) :: acc)
        infos.lib_units
        tolink

let make_startup_class units_to_link execname =
  let class_name = !Jclflags.java_package ^ ".cafesterolMain" in
  let main_class = make_class class_name in
  let classes =
    List.flatten
      (List.map
         (fun (infos, _, _) ->
           infos.ui_additional_classes @ [infos.ui_javaname])
         units_to_link) in
  let constructor_code =
    [ Instruction.ALOAD_0 ;
      Instruction.ALOAD_1 ]
    @ cstr_AbstractNativeRunner
    @ [ Instruction.RETURN ] in
  let constructor_method =
    Method.Constructor ([`Private],
                        [`Class class_NativeParameters],
                        [`Code { Attribute.max_stack = 2;
                                 Attribute.max_locals = 2;
                                 Attribute.code = constructor_code;
                                 Attribute.exception_table = [];
                                 Attribute.attributes = []; }]) in
  let copy_constructor_code =
    [ Instruction.ALOAD_0 ;
      Instruction.ALOAD_1 ]
    @ cstr_copy_AbstractNativeRunner
    @ [ Instruction.RETURN ] in
  let copy_constructor_method =
    Method.Constructor ([`Private],
                        [`Class main_class],
                        [`Code { Attribute.max_stack = 2;
                                 Attribute.max_locals = 2;
                                 Attribute.code = copy_constructor_code;
                                 Attribute.exception_table = [];
                                 Attribute.attributes = []; }]) in
  let copy_code =
    [ Instruction.NEW main_class ;
      Instruction.DUP ;
      Instruction.ALOAD_0 ;
      Instruction.INVOKESPECIAL (main_class,
                                 make_method "<init>",
                                 ([`Class main_class], `Void)) ;
      Instruction.ARETURN ] in
  let copy_method =
    Method.Regular ([`Public],
                    (make_method "copy"),
                    ([], `Class class_AbstractNativeRunner),
                    [`Code { Attribute.max_stack = 3;
                             Attribute.max_locals = 1;
                             Attribute.code = copy_code;
                             Attribute.exception_table = [];
                             Attribute.attributes = []; }]) in
  let module_main_code =
    (List.concat
       (List.map
          (fun c ->
            [ Instruction.ALOAD_0 ;
              Instruction.DUP ;
              Instruction.DUP ;
              Instruction.INVOKESTATIC
                (make_class c,
                 make_method "entry",
                 ([`Class class_AbstractNativeRunner], `Class class_Value)) ;
              Instruction.PUTFIELD (class_AbstractNativeRunner,
                                    make_field "result",
                                    `Class class_Value) ;
              Instruction.INVOKEVIRTUAL
                (`Class_or_interface class_AbstractNativeRunner,
                 make_method "incrGlobalsInited",
                 ([], `Void)) ])
          classes))
    @ [ Instruction.RETURN ] in
  let module_main_method =
    Method.Regular ([`Protected],
                    (make_method "moduleMain"),
                    ([], `Void),
                    [`Code { Attribute.max_stack = 3;
                             Attribute.max_locals = 1;
                             Attribute.code = module_main_code;
                             Attribute.exception_table = [];
                             Attribute.attributes = []; }]) in
  let main_with_return_code =
    [ Instruction.NEW main_class ;
      Instruction.DUP ;

      (* parameters *)
      Instruction.LDC_W (`Class_or_interface main_class) ;
      Instruction.LDC_W (`String (utf8_of_string "cadmium.parameters")) ]
    @ meth_getResourceAsStream
    @ [ Instruction.ALOAD_0 ]
    @ field_IN
    @ field_OUT
    @ field_ERR
    @ [ (if !Jclflags.applet then Instruction.ICONST_1 else Instruction.ICONST_0) ;
        Instruction.LDC_W (`String (utf8_of_string execname)) ]
    @ meth_fromStream
    @ [ Instruction.INVOKESPECIAL
          (main_class,
           make_method "<init>",
           ([`Class class_NativeParameters], `Void)) ;
        Instruction.ASTORE_1 ]
    @ (List.concat
         (List.map
            (fun e ->
              [ Instruction.ALOAD_1 ;
                Instruction.LDC_W (`String (utf8_of_string e)) ]
              @ meth_registerPredefinedException)
            (Array.to_list Runtimedef.builtin_exceptions)))
    @ (List.concat
         (List.map
            (fun c ->
              [ Instruction.ALOAD_1 ;
                Instruction.LDC_W (`Class_or_interface (make_class c)) ]
              @ meth_loadConstant)
            classes))
    @ [  Instruction.ALOAD_1 ]
    @ meth_execute
    @ [ Instruction.ALOAD_1 ;
        Instruction.ARETURN ] in
  let main_with_return_method =
    Method.Regular ([`Public; `Static],
                    (make_method "mainWithReturn"),
                    ([`Array (`Class class_String)], `Class main_class),
                    [`Code { Attribute.max_stack = 9;
                             Attribute.max_locals = 2;
                             Attribute.code = main_with_return_code;
                             Attribute.exception_table = [];
                             Attribute.attributes = []; }]) in
  let main_scripting_code =
    [ Instruction.NEW main_class ;
      Instruction.DUP ;

      (* parameters *)
      Instruction.LDC_W (`Class_or_interface main_class) ;
      Instruction.LDC_W (`String (utf8_of_string "cadmium.parameters")) ]
    @ meth_getResourceAsStream
    @ [ Instruction.ALOAD_0 ;
        Instruction.ALOAD_2 ;
        Instruction.ALOAD_3 ;
        Instruction.ALOAD 4 ;
        Instruction.ICONST_0 ;
        Instruction.LDC_W (`String (utf8_of_string execname)) ]
    @ meth_fromStream
    @ [ Instruction.INVOKESPECIAL
          (main_class,
           make_method "<init>",
           ([`Class class_NativeParameters], `Void)) ;
        Instruction.ASTORE 5 ]
    @ (List.concat
         (List.map
            (fun e ->
              [ Instruction.ALOAD 5 ;
                Instruction.LDC_W (`String (utf8_of_string e)) ]
              @ meth_registerPredefinedException)
            (Array.to_list Runtimedef.builtin_exceptions)))
    @ (List.concat
         (List.map
            (fun c ->
              [ Instruction.ALOAD 5 ;
                Instruction.LDC_W (`Class_or_interface (make_class c)) ]
              @ meth_loadConstant)
            classes))
    @ [  Instruction.ALOAD 5 ;
         Instruction.ALOAD_1 ]
    @ meth_executeWithBindings
    @ [  Instruction.ALOAD 5 ]
    @ meth_getResult
    @ [ Instruction.ARETURN ] in
  let main_scripting_method =
    Method.Regular ([`Public; `Static],
                    (make_method "mainScripting"),
                    ([`Array (`Class class_String);
                      `Class class_Map;
                      `Class class_InputStream;
                      `Class class_PrintStream;
                      `Class class_PrintStream], `Class class_Value),
                    [`Code { Attribute.max_stack = 9;
                             Attribute.max_locals = 6;
                             Attribute.code = main_scripting_code;
                             Attribute.exception_table = [];
                             Attribute.attributes = []; }]) in
  let main_code =
    [ Instruction.ALOAD_0 ;
      Instruction.INVOKESTATIC
        (main_class,
         make_method "mainWithReturn",
         ([`Array (`Class class_String)], `Class main_class)) ;
      Instruction.POP ;
      Instruction.RETURN ] in
  let main_method =
    Method.Regular ([`Public; `Static],
                    (make_method "main"),
                    ([`Array (`Class class_String)], `Void),
                    [`Code { Attribute.max_stack = 1;
                             Attribute.max_locals = 1;
                             Attribute.code = main_code;
                             Attribute.exception_table = [];
                             Attribute.attributes = []; }]) in
  let cd = { ClassDefinition.access_flags = [`Public; `Final; `Super];
             name = main_class;
             extends = Some class_AbstractNativeRunner;
             implements = [];
             fields = [];
             methods = [ constructor_method ;
                         copy_constructor_method ;
                         copy_method ;
                         module_main_method ;
                         main_with_return_method ;
                         main_scripting_method ;
                         main_method ];
             attributes = []; } in
  let buff = Buffer.create 16 in
  let os = OutputStream.make_of_buffer buff in
  ClassFile.write (ClassDefinition.encode cd) os;
  OutputStream.close os;
  class_name, (Buffer.contents buff)

let make_applet_class () =
  let main_class_name = !Jclflags.java_package ^ ".cafesterolMain" in
  let main_class = make_class main_class_name in
  let class_name = !Jclflags.java_package ^ ".cafesterolApplet" in
  let constructor_code =
    [ Instruction.ALOAD_0 ]
    @ cstr_AbstractNativeApplet
    @ [ Instruction.ALOAD_0 ;
        Instruction.ICONST_0 ;
        Instruction.ANEWARRAY (`Class_or_interface class_String) ;
        Instruction.INVOKESTATIC
          (main_class,
           make_method "mainWithReturn",
           ([`Array (`Class class_String)], `Class main_class)) ]
    @ field_applet_runner
    @ [ Instruction.RETURN ] in
  let constructor_method =
    Method.Constructor ([`Public],
                        [],
                        [`Code { Attribute.max_stack = 2;
                                 Attribute.max_locals = 1;
                                 Attribute.code = constructor_code;
                                 Attribute.exception_table = [];
                                 Attribute.attributes = []; }]) in
  let cd = { ClassDefinition.access_flags = [`Public; `Final; `Super];
             name = make_class class_name;
             extends = Some class_AbstractNativeApplet;
             implements = [];
             fields = [];
             methods = [ constructor_method ];
             attributes = []; } in
  let buff = Buffer.create 16 in
  let os = OutputStream.make_of_buffer buff in
  ClassFile.write (ClassDefinition.encode cd) os;
  OutputStream.close os;
  class_name, (Buffer.contents buff)

let make_servlet_class () =
  let main_class_name = !Jclflags.java_package ^ ".cafesterolMain" in
  let main_class = make_class main_class_name in
  let class_name = !Jclflags.java_package ^ ".cafesterolServlet" in
  let constructor_code =
    [ Instruction.ALOAD_0 ]
    @ cstr_AbstractNativeServlet
    @ [ Instruction.ALOAD_0 ;
        Instruction.ICONST_0 ;
        Instruction.ANEWARRAY (`Class_or_interface class_String) ;
        Instruction.INVOKESTATIC
          (main_class,
           make_method "mainWithReturn",
           ([`Array (`Class class_String)], `Class main_class)) ]
    @ [ Instruction.SWAP ;
        Instruction.POP ]
    @ field_servlet_runner
    @ [ Instruction.RETURN ] in
  let constructor_method =
    Method.Constructor ([`Public],
                        [],
                        [`Code { Attribute.max_stack = 2;
                                 Attribute.max_locals = 1;
                                 Attribute.code = constructor_code;
                                 Attribute.exception_table = [];
                                 Attribute.attributes = []; }]) in
  let cd = { ClassDefinition.access_flags = [`Public; `Final; `Super];
             name = make_class class_name;
             extends = Some class_AbstractNativeServlet;
             implements = [];
             fields = [];
             methods = [ constructor_method ];
             attributes = []; } in
  let buff = Buffer.create 16 in
  let os = OutputStream.make_of_buffer buff in
  ClassFile.write (ClassDefinition.encode cd) os;
  OutputStream.close os;
  class_name, (Buffer.contents buff)

let link objfiles filename =
  let entries = Hashtbl.create 16 in
  let check_entry s =
    if Hashtbl.mem entries s then
      raise (Error (Duplicate_entry s))
    else
      Hashtbl.add entries s () in
  let internal_name s =
    let s' = String.copy s in
    for i = 0 to pred (String.length s') do
      if s'.[i] = '.' then s'.[i] <- '/'
    done;
    match !Jclflags.servlet with
    | Some _ -> "WEB-INF/classes/" ^ s'
    | None -> s' in
  let objfiles =
    if !Clflags.nopervasives then
      objfiles
    else
      "stdlib.cmja" :: objfiles in
  let units_to_link = List.fold_right scan_file objfiles [] in
  Array.iter remove_required Runtimedef.builtin_exceptions;
  (match extract_missing_globals () with
  | [] -> ()
  | mg -> raise (Error (Missing_implementations mg)));
  List.iter
    (fun (info, filename, crc) -> check_consistency filename info crc)
    units_to_link;
  let startup_name, startup_class =
    make_startup_class units_to_link (Filename.basename filename) in
  try
    let zip = Zip.open_out filename in
    let cname = internal_name startup_name ^ ".class" in
    let gname = internal_name startup_name ^ ".gmap" in
    check_entry cname;
    check_entry gname;
    Zip.add_entry startup_class zip cname;
    let gmap =
      List.map
        (fun (unit, _, crc) ->
          try
            (unit.ui_name, (List.assoc unit.ui_name unit.ui_imports_cmi), crc, unit.ui_defines)
          with Not_found -> assert false)
        units_to_link in
    Zip.add_entry (Marshal.to_string gmap []) zip gname;
    if !Jclflags.applet then begin
      let applet_name, applet_class = make_applet_class () in
      let aname = internal_name applet_name ^ ".class" in
      check_entry aname;
      Zip.add_entry applet_class zip aname;
    end;
    (match !Jclflags.servlet with
    | Some file ->
        let servlet_name, servlet_class = make_servlet_class () in
        let sname = internal_name servlet_name ^ ".class" in
        check_entry sname;
        Zip.add_entry servlet_class zip sname;
        (try
          check_entry "WEB-INF/web.xml";
          Zip.copy_file_to_entry file zip "WEB-INF/web.xml"
        with _ -> raise (Error (File_not_found file)))
    | None -> ());
    let cmja_files =
      List.filter (fun s -> Filename.check_suffix s ".cmja") objfiles in
    let jars = List.map
        (fun s -> (Filename.chop_suffix s ".cmja") ^ Jconfig.ext_lib)
        cmja_files in
    let cp =
      if cmja_files = [] then
        ""
      else
        " " ^ (String.concat " " jars) in
    let manifest = String.concat "\n" ([
      "Manifest-Version: 1.0" ;
      "Main-Class: " ^ startup_name ]
      @ (if !Jclflags.standalone then
        [ "" ]
      else
        [ "Class-Path: ocamlrun.jar"
          ^ cp
          ^ (if !Jclflags.additional_jars = [] then
            ""
          else
            (String.concat " " !Jclflags.additional_jars)) ;
          "" ])) in
    check_entry "META-INF/MANIFEST.MF";
    Zip.add_entry manifest zip "META-INF/MANIFEST.MF";
    let params = String.concat "\n" (List.rev !Jclflags.cadmium_parameters) in
    let pname = internal_name !Jclflags.java_package ^ "/cadmium.parameters" in
    check_entry pname;
    Zip.add_entry params zip pname;
    let cmj_files =
      List.filter (fun s -> Filename.check_suffix s ".cmj") objfiles in
    List.iter
      (fun fn ->
        let fn' =
          try
            Misc.find_in_path
              !Config.load_path
              (Filename.chop_suffix fn ".cmj" ^ Jconfig.ext_obj)
          with Not_found -> raise (Error (File_not_found fn)) in
        let z = Zip.open_in fn' in
        List.iter
          (fun e ->
            if not e.Zip.is_directory then
              let data = Zip.read_entry z e in
              let name = match !Jclflags.servlet with
              | Some _ -> "WEB-INF/classes/" ^ e.Zip.filename
              | None -> e.Zip.filename in
              check_entry name;
              Zip.add_entry data zip name)
          (Zip.entries z);
        Zip.close_in z)
      cmj_files;
    if !Jclflags.standalone then begin
      let service_prefix = "META-INF/services/" in
      let service_prefix_length = String.length service_prefix in
      let services = Hashtbl.create 16 in
      List.iter
        (fun jar ->
          let fn =
            try
              Misc.find_in_path !Config.load_path jar
            with Not_found -> raise (Error (File_not_found jar)) in
          match !Jclflags.servlet with
          | None ->
              let z = Zip.open_in fn in
              List.iter
                (fun e ->
                  let fname = e.Zip.filename in
                  if not (e.Zip.is_directory) then begin
                    if ((String.length fname) >= service_prefix_length)
                        && (String.sub fname 0 service_prefix_length) = service_prefix then
                      let l = try Hashtbl.find services fname with Not_found -> [] in
                      let l' = (Zip.read_entry z e) :: l in
                      Hashtbl.replace services fname l'
                    else if (fname <> "META-INF/MANIFEST.MF") then
                      let data = Zip.read_entry z e in
                      check_entry fname;
                      Zip.add_entry data zip fname
                    else
                      ()
                  end)
                (Zip.entries z);
              Zip.close_in z
          | Some _ ->
              let lname = "WEB-INF/lib/" ^ (Filename.basename fn) in
              check_entry lname;
              Zip.copy_file_to_entry fn zip lname)
        ("ocamlrun.jar" :: jars @ !Jclflags.additional_jars);
      Hashtbl.iter
        (fun name l ->
          if !Jclflags.no_merge && ((List.length l) > 1) then
            raise (Error (Duplicate_entry name))
          else
            let data = Buffer.create 16 in
            List.iter
              (fun d ->
                if (Buffer.length data) > 0 then Buffer.add_char data '\n' else ();
                Buffer.add_string data d)
              (List.rev l);
            Zip.add_entry (Buffer.contents data) zip name)
        services
    end;
    List.iter
      (fun f ->
        try
          let is = InputStream.make_of_channel (open_in_bin f) in
          let cf = ClassDefinition.decode (ClassFile.read is) in
          let name = string_of_utf8 (Name.internal_utf8_for_class cf.ClassDefinition.name) in
          check_entry (name ^ ".class");
          Zip.copy_file_to_entry f zip (name ^ ".class")
        with _ -> raise (Error (Unable_to_add_class f)))
      !Jclflags.additional_classes;
    List.iter
      (fun f ->
        try
          let file, path = try
            let idx = String.index f ':' in
            (String.sub f 0 idx), (String.sub f (succ idx) ((String.length f) - idx - 1))
          with Not_found -> f, "" in
          let base = Filename.basename file in
          let name = if path = "" then base else path ^ "/" ^ base in
          check_entry name;
          Zip.copy_file_to_entry file zip name
        with _ -> raise (Error (Unable_to_add_file f)))
      !Jclflags.additional_files;
    Zip.close_out zip
  with
  | (Error _) as e -> raise e
  | _ -> raise (Error Linking_error)

type dynunit = {
    name : string;
    crc : Digest.t;
    imports_cmi : (string * Digest.t) list;
    imports_cmj : (string * Digest.t) list;
    defines : string list;
  }

type dynheader = {
    magic : string;
    units : dynunit list;
  }

let plugin_magic_number = "Caml2007D001"

let plugin_header units =
  let conv (ui, crc) =
    { name = ui.Jcompilenv.ui_name;
      crc = crc;
      imports_cmi = ui.Jcompilenv.ui_imports_cmi;
      imports_cmj = ui.Jcompilenv.ui_imports_cmj;
      defines = ui.Jcompilenv.ui_defines } in
  let h = { magic = plugin_magic_number;
            units = List.map conv units } in
  Marshal.to_string h []

let plugin_members units =
  let buff = Buffer.create 32 in
  List.iter
    (fun (ui, _) ->
      Buffer.add_string buff ui.Jcompilenv.ui_name;
      Buffer.add_char buff '=';
      Buffer.add_string buff ui.Jcompilenv.ui_javaname;
      Buffer.add_char buff '\n';
      if ui.Jcompilenv.ui_additional_classes <> [] then begin
        Buffer.add_string buff ui.Jcompilenv.ui_name;
        Buffer.add_string buff "*=";
        Buffer.add_string buff (String.concat "," ui.Jcompilenv.ui_additional_classes);
        Buffer.add_char buff '\n'
      end)
    units;
  Buffer.contents buff

let link_shared objfiles filename =
  let entries = Hashtbl.create 16 in
  let check_entry s =
    if Hashtbl.mem entries s then
      raise (Error (Duplicate_entry s))
    else
      Hashtbl.add entries s () in
  let units_to_link = List.fold_right scan_file objfiles [] in
  List.iter
    (fun (info, filename, crc) -> check_consistency filename info crc)
    units_to_link;
  let units = List.map (fun (ui, _, crc) -> (ui, crc)) units_to_link in
  let header = plugin_header units in
  let members = plugin_members units in
  try
    let zip = Zip.open_out filename in
    Zip.add_entry header zip "/PluginHeader";
    Zip.add_entry members zip "/PluginMembers";
    List.iter
      (fun fn ->
        let fn' =
          try
            Misc.find_in_path
              !Config.load_path
              (Filename.chop_suffix fn ".cmj" ^ Jconfig.ext_obj)
          with Not_found -> raise (Error (File_not_found fn)) in
        let z = Zip.open_in fn' in
        List.iter
          (fun e ->
            if not e.Zip.is_directory then
              let data = Zip.read_entry z e in
              let name = e.Zip.filename in
              check_entry name;
              Zip.add_entry data zip name)
          (Zip.entries z);
        Zip.close_in z)
      (List.filter (fun s -> Filename.check_suffix s ".cmj") objfiles);
    Zip.close_out zip
  with
  | (Error _) as e -> raise e
  | _ -> raise (Error Linking_error)

let report_error ppf = function
  | File_not_found name ->
      Format.fprintf ppf "Cannot find file %s" name
  | Not_an_object_file name ->
      Format.fprintf ppf "The file %s is not a compilation unit description" name
  | Missing_implementations l ->
     let print_references ppf = function
       | [] -> ()
       | r1 :: rl ->
           Format.fprintf ppf "%s" r1;
           List.iter (fun r -> Format.fprintf ppf ",@ %s" r) rl in
      let print_modules ppf =
        List.iter
         (fun (md, rq) ->
            Format.fprintf ppf "@ @[<hov 2>%s referenced from %a@]" md
            print_references rq) in
      Format.fprintf ppf
       "@[<v 2>No implementations provided for the following modules:%a@]"
       print_modules l
  | Inconsistent_interface(intf, file1, file2) ->
      Format.fprintf ppf
       "@[<hov>Files %s@ and %s@ make inconsistent assumptions \
              over interface %s@]"
       file1 file2 intf
  | Inconsistent_implementation(intf, file1, file2) ->
      Format.fprintf ppf
       "@[<hov>Files %s@ and %s@ make inconsistent assumptions \
              over implementation %s@]"
       file1 file2 intf
  | Linking_error ->
      Format.fprintf ppf "Error during linking"
  | Multiple_definition(modname, file1, file2) ->
      Format.fprintf ppf
        "@[<hov>Files %s@ and %s@ both define a module named %s@]"
        file1 file2 modname
  | Missing_cmj(filename, name) ->
      Format.fprintf ppf
        "@[<hov>File %s@ was compiled without access@ \
         to the .cmj file@ for module %s,@ \
         which was produced by `ocamljava -for-pack'.@ \
         Please recompile %s@ with the correct `-I' option@ \
         so that %s.cmj@ is found.@]"
        filename name filename name
  | Unable_to_add_class filename ->
      Format.fprintf ppf "Unable to add class from file %s" filename
  | Unable_to_add_file filename ->
      Format.fprintf ppf "Unable to add file %s" filename
  | Duplicate_entry name ->
      Format.fprintf ppf "Duplicate jar entry (%s)" name
