(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2010 Xavier Clerc.
 * Original file (asmcomp/closure.ml in the Objective Caml source
 * distribution) is Copyright (C) INRIA.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://qt.nokia.com/doc/4.0/qpl.html>.
 *)

open Lambda
open Asttypes
open Jlambda


let rec split_list n l =
  if n <= 0 then
    ([], l)
  else
    match l with
    | [] -> Misc.fatal_error "Jlambda.split_list"
    | a :: l -> let l1, l2 = split_list (pred n) l in (a :: l1, l2)

let rec build_closure_env env_param pos = function
  | [] -> Tbl.empty
  | id :: tail ->
      Tbl.add
        id
        (Jprim (Pfield pos, [Jvar env_param], Debuginfo.none))
        (build_closure_env env_param (succ pos) tail)

let getglobal id =
  Jprim (Pgetglobal (Ident.create_persistent (Jcompilenv.symbol_for_global id)),
         [],
         Debuginfo.none)

let occurs_var var j =
  let rec occurs = function
    | Jvar v -> v = var
    | Jconst _ -> false
    | Jdirect_apply (_, args, _) -> List.exists occurs args
    | Jgeneric_apply (func, args, _) -> occurs func || List.exists occurs args
    | Jclosure (_, clos) -> List.exists occurs clos
    | Joffset (j, _) -> occurs j
    | Jlet (_, def, body) -> occurs def || occurs body
    | Jletrec (decls, body) -> List.exists (fun (_, j) -> occurs j) decls || occurs body
    | Jprim (_, args, _) -> List.exists occurs args
    | Jswitch (arg, s) -> occurs arg || occurs_array s.js_actions_consts || occurs_array s.js_actions_blocks
    | Jstaticfail (_, args) -> List.exists occurs args
    | Jstaticcatch (_, _, body, handler) -> occurs body || occurs handler
    | Jtrywith (body, _, handler) -> occurs body || occurs handler
    | Jifthenelse (cond, ifso, ifno) -> occurs cond || occurs ifso || occurs ifno
    | Jsequence (j1, j2) -> occurs j1 || occurs j2
    | Jwhile (cond, body) -> occurs cond || occurs body
    | Jfor (_, lo, hi, _, body) -> occurs lo || occurs hi || occurs body
    | Jassign (id, j) -> id = var || occurs j
    | Jsend (_, meth, obj, args, _) -> occurs meth || occurs obj || List.exists occurs args
  and occurs_array a =
    try
      for i = 0 to pred (Array.length a) do
        if occurs a.(i) then raise Exit
      done;
      false
    with Exit ->
      true in
  occurs j

let prim_size prim args =
  match prim with
  | Pidentity -> 0
  | Pgetglobal _ -> 1
  | Psetglobal _ -> 1
  | Pmakeblock _ -> 5 + List.length args
  | Pfield _ -> 1
  | Psetfield (_, is_ptr) -> if is_ptr then 4 else 1
  | Pfloatfield _ -> 1
  | Psetfloatfield _ -> 1
  | Pduprecord _ -> 10 + List.length args
  | Pccall p -> (if p.Primitive.prim_alloc then 10 else 4) + List.length args
  | Praise -> 4
  | Pstringlength -> 5
  | Pstringrefs -> 6
  | Pstringsets -> 6
  | Pmakearray _ -> 5 + List.length args
  | Parraylength kind -> if kind = Pgenarray then 6 else 2
  | Parrayrefu kind -> if kind = Pgenarray then 12 else 2
  | Parraysetu kind -> if kind = Pgenarray then 16 else 4
  | Parrayrefs kind -> if kind = Pgenarray then 18 else 8
  | Parraysets kind -> if kind = Pgenarray then 22 else 10
  | Pbittest -> 3
  | Pbigarrayref (_, ndims, _, _) -> 4 + ndims * 6
  | Pbigarrayset (_, ndims, _, _) -> 4 + ndims * 6
  | _ -> 2

let lambda_smaller lam threshold =
  let size = ref 0 in
  let rec lambda_size lam =
    if !size > threshold then raise Exit;
    match lam with
    | Jvar _ -> ()
    | Jconst (Const_base (Const_int _ | Const_char _ | Const_float _ | Const_int32 _ | Const_int64 _ | Const_nativeint _) | Const_pointer _) -> incr size
    | Jconst _ -> raise Exit
    | Jdirect_apply (_, args, _) ->
        size := !size + 4;
        lambda_list_size args
    | Jgeneric_apply (func, args, _) ->
        size := !size + 6;
        lambda_size func;
        lambda_list_size args
    | Jclosure _ -> raise Exit
    | Joffset (lam, _) ->
        incr size;
        lambda_size lam
    | Jlet (_, lam, body) ->
        lambda_size lam;
        lambda_size body
    | Jletrec _ -> raise Exit
    | Jprim (prim, args, _) ->
        size := !size + prim_size prim args;
        lambda_list_size args
    | Jswitch (lam, cases) ->
        if Array.length cases.js_actions_consts > 1 then size := !size + 5;
        if Array.length cases.js_actions_blocks > 1 then size := !size + 5;
        lambda_size lam;
        lambda_array_size cases.js_actions_consts;
        lambda_array_size cases.js_actions_blocks
    | Jstaticfail (_, args) ->
        lambda_list_size args
    | Jstaticcatch (_, _, body, handler) ->
        incr size;
        lambda_size body;
        lambda_size handler
    | Jtrywith (body, _, handler) ->
        size := !size + 8;
        lambda_size body;
        lambda_size handler
    | Jifthenelse (cond, ifso, ifno) ->
        size := !size + 2;
        lambda_size cond;
        lambda_size ifso;
        lambda_size ifno
    | Jsequence (j1, j2) ->
        lambda_size j1;
        lambda_size j2
    | Jwhile (cond, body) ->
        size := !size + 2;
        lambda_size cond;
        lambda_size body
    | Jfor (_, lo, hi, _, body) ->
        size := !size + 4;
        lambda_size lo;
        lambda_size hi;
        lambda_size body
    | Jassign (_, lam) ->
        incr size;
        lambda_size lam
    | Jsend (_, meth, obj , args, _) ->
        size := !size + 8;
        lambda_size meth;
        lambda_size obj;
        lambda_list_size args
  and lambda_list_size l = List.iter lambda_size l
  and lambda_array_size a = Array.iter lambda_size a in
  try
    lambda_size lam;
    !size <= threshold
  with Exit ->
    false

let rec is_pure_jlambda = function
  | Jvar _ -> true
  | Jconst _ -> true
  | Jprim (p, args, _) ->
      (match p with
      | Psetglobal _
      | Psetfield _
      | Psetfloatfield _
      | Pduprecord _
      | Pccall _
      | Praise
      | Poffsetref _
      | Pstringsetu
      | Pstringsets
      | Parraysetu _
      | Parraysets _
      | Pbigarrayset _ -> false
      | _ -> List.for_all is_pure_jlambda args)
  | _ -> false

let make_const_int n = (Jconst (Const_base (Const_int n))), (Value_integer n)
let make_const_ptr n = (Jconst (Const_pointer n)), (Value_constptr n)
let make_const_bool b = make_const_ptr (if b then 1 else 0)

let simplif_prim_pure p (args, approxs) dbg =
  match approxs with
  | [Value_integer x] ->
      (match p with
      | Pidentity -> make_const_int x
      | Pnegint -> make_const_int (-x)
      | Poffsetint y -> make_const_int (x + y)
      | _ -> (Jprim (p, args, dbg)), Value_unknown)
  | [Value_integer x; Value_integer y] ->
      (match p with
      | Paddint -> make_const_int (x + y)
      | Psubint -> make_const_int (x - y)
      | Pmulint -> make_const_int (x * y)
      | Pdivint when y <> 0 -> make_const_int (x / y)
      | Pmodint when y <> 0 -> make_const_int (x mod y)
      | Pandint -> make_const_int (x land y)
      | Porint -> make_const_int (x lor y)
      | Pxorint -> make_const_int (x lxor y)
      | Plslint -> make_const_int (x lsl y)
      | Plsrint -> make_const_int (x lsr y)
      | Pasrint -> make_const_int (x asr y)
      | Pintcomp cmp ->
          let res = match cmp with
          | Ceq -> x = y
          | Cneq -> x <> y
          | Clt -> x < y
          | Cgt -> x > y
          | Cle -> x <= y
          | Cge -> x >= y in
          make_const_bool res
      | _ -> (Jprim (p, args, dbg)), Value_unknown)
  | [Value_constptr x] ->
      (match p with
      | Pidentity -> make_const_ptr x
      | Pnot -> make_const_bool (x = 0)
      | Pisint -> make_const_bool true
      | _ -> (Jprim (p, args, dbg)), Value_unknown)
  | [Value_constptr x; Value_constptr y] ->
      (match p with
      | Psequand -> make_const_bool (x <> 0 && y <> 0)
      | Psequor -> make_const_bool (x <> 0 || y <> 0)
      | _ -> (Jprim (p, args, dbg)), Value_unknown)
  | _ -> (Jprim (p, args, dbg)), Value_unknown

let simplif_prim p (args, _ as args_approxs) dbg =
  if List.for_all is_pure_jlambda args then
    simplif_prim_pure p args_approxs dbg
  else
    (Jprim (p, args, dbg)), Value_unknown

let approx_jlam = function
  | Jconst (Const_base (Const_int n)) -> Value_integer n
  | Jconst (Const_base (Const_char c)) -> Value_integer (Char.code c)
  | Jconst (Const_pointer n) -> Value_constptr n
  | _ -> Value_unknown

let rec substitute sb jlam =
  match jlam with
  | Jvar v -> (try Tbl.find v sb with Not_found -> jlam)
  | Jconst _ -> jlam
  | Jdirect_apply (label, args, dbg) ->
      Jdirect_apply (label, List.map (substitute sb) args, dbg)
  | Jgeneric_apply (func, args, dbg) ->
      Jgeneric_apply (substitute sb func, List.map (substitute sb) args, dbg)
  | Jclosure (defs, env) -> Jclosure (defs, List.map (substitute sb) env)
  | Joffset (j, ofs) -> Joffset (substitute sb j, ofs)
  | Jlet (id, j1, j2) ->
      let id' = Ident.rename id in
      Jlet (id', substitute sb j1, substitute (Tbl.add id (Jvar id') sb) j2)
  | Jletrec (bindings, body) ->
      let bindings' = List.map
          (fun (id, x) -> (id, Ident.rename id, x))
          bindings in
      let sb' = List.fold_right
          (fun (id, id', _) s -> Tbl.add id (Jvar id') s)
          bindings'
          sb in
      Jletrec (List.map (fun (_, id', x) -> (id', substitute sb' x)) bindings',
               substitute sb' body)
  | Jprim (p, args, dbg) ->
      let args' = List.map (substitute sb) args in
      fst (simplif_prim p (args', List.map approx_jlam args') dbg)
  | Jswitch (arg, sw) ->
      Jswitch (substitute sb arg,
               { sw with
                 js_actions_consts = Array.map (substitute sb) sw.js_actions_consts;
                 js_actions_blocks = Array.map (substitute sb) sw.js_actions_blocks; })
  | Jstaticfail (fail, args) -> Jstaticfail (fail, List.map (substitute sb) args)
  | Jstaticcatch (fail, ids, j1, j2) ->
      Jstaticcatch (fail, ids, substitute sb j1, substitute sb j2)
  | Jtrywith (j1, id, j2) ->
      let id' = Ident.rename id in
      Jtrywith (substitute sb j1,
                id',
                substitute (Tbl.add id (Jvar id') sb) j2)
  | Jifthenelse (j1, j2, j3) ->
      (match substitute sb j1 with
      | Jconst (Const_pointer n) ->
          if n <> 0 then substitute sb j2 else substitute sb j3
      | j1' -> Jifthenelse (j1', substitute sb j2, substitute sb j3))
  | Jsequence (j1, j2) -> Jsequence (substitute sb j1, substitute sb j2)
  | Jwhile (j1, j2) -> Jwhile (substitute sb j1, substitute sb j2)
  | Jfor (id, j1, j2, dir, j3) ->
      let id' = Ident.rename id in
      Jfor (id',
            substitute sb j1,
            substitute sb j2,
            dir,
            substitute (Tbl.add id (Jvar id') sb) j3)
  | Jassign (id, j) ->
      let id' = try
        match Tbl.find id sb with Jvar i -> i | _ -> assert false
      with Not_found -> id in
      Jassign (id', substitute sb j)
  | Jsend (kind, j1, j2, jl, dbg) ->
      Jsend (kind,
             substitute sb j1,
             substitute sb j2,
             List.map (substitute sb) jl,
             dbg)

let is_simple_argument = function
  | Jvar _ -> true
  | Jconst (Const_base (Const_int _ | Const_char _ | Const_float _ | Const_int32 _ | Const_int64 _ | Const_nativeint _)) -> true
  | Jconst (Const_pointer _) -> true
  | _ -> false

let no_effects = function
  | Jclosure _ -> true
  | Jconst (Const_base (Const_string _)) -> true
  | x -> is_simple_argument x

let rec bind_params_rec subst params args body =
  match params, args with
  | [], [] -> substitute subst body
  | (p :: pl, a :: al) ->
      if is_simple_argument a then
        bind_params_rec (Tbl.add p a subst) pl al body
      else begin
        let p' = Ident.rename p in
        let body' = bind_params_rec (Tbl.add p (Jvar p') subst) pl al body in
        if occurs_var p body then
          Jlet (p', a, body')
        else if no_effects a then
          body'
        else
          Jsequence (a, body')
      end
  | _ -> assert false

let bind_params params args body =
  bind_params_rec Tbl.empty (List.rev params) (List.rev args) body

let rec is_pure = function
  | Lvar _ -> true
  | Lconst _ -> true
  | Lprim (p, args) ->
      (match p with
      | Psetglobal _
      | Psetfield _
      | Psetfloatfield _
      | Pduprecord _
      | Pccall _
      | Praise
      | Poffsetref _
      | Pstringsetu
      | Pstringsets
      | Parraysetu _
      | Parraysets _
      | Pbigarrayset _ -> false
      | _ -> List.for_all is_pure args)
  | _ -> false

let direct_apply fundesc func jfunc jargs =
  let app_args = if fundesc.fun_closed then jargs else jargs @ [jfunc] in
  let app = match fundesc.fun_inline with
  | None -> Jdirect_apply (fundesc.fun_label, app_args, Debuginfo.none)
  | Some (params, body) -> bind_params params app_args body in
  if not fundesc.fun_closed || is_pure func then
    app
  else
    Jsequence (jfunc, app)

let strengthen_approx app approx =
  match approx_jlam app with
  | (Value_integer _ | Value_constptr _) as x -> x
  | _ -> approx

let check_constant_result lam jlam approx =
  match approx with
  | Value_integer n when is_pure lam -> make_const_int n
  | Value_constptr n when is_pure lam -> make_const_ptr n
  | _ -> jlam, approx

let sequence_constant_expr lam jlam1 (jlam2, approx2 as res2) =
  if is_pure lam then
    res2
  else
    (Jsequence (jlam1, jlam2)), approx2

let global_approx = ref ([||] : value_approximation array)

let depth = ref 0
let excessive_depth = 5

let rec add_debug_info ev j =
  match ev.lev_kind with
  | Lev_after _ ->
      (match j with
      | Jdirect_apply (lbl, args, _) ->
          Jdirect_apply (lbl, args, Debuginfo.from_call ev)
      | Jgeneric_apply (Jdirect_apply (lbl, args1, _), args2, _) ->
          Jgeneric_apply (Jdirect_apply (lbl, args1, Debuginfo.from_call ev),
                          args2,
                          Debuginfo.from_call ev)
      | Jgeneric_apply (func, args, _) ->
          Jgeneric_apply (func, args, Debuginfo.from_call ev)
      | Jprim (Praise, args, _) ->
          Jprim (Praise, args, Debuginfo.from_call ev)
      | Jprim (p, args, _) ->
          Jprim (p, args, Debuginfo.from_call ev)
      | Jsend (kind, obj, meth, args, _) ->
          Jsend (kind, obj, meth, args, Debuginfo.from_call ev)
      | Jsequence (j1, j2) ->
          Jsequence (j1, add_debug_info ev j2)
      | _ -> j)
  | _ -> j

let close_approx_var fenv cenv id =
  let approx = try Tbl.find id fenv with Not_found -> Value_unknown in
  match approx with
  | Value_integer n -> make_const_int n
  | Value_constptr n -> make_const_ptr n
  | approx ->
      let subst = try Tbl.find id cenv with Not_found -> Jvar id in
      (subst, approx)

let close_var fenv cenv id =
  fst (close_approx_var fenv cenv id)

let rec close fenv cenv = function
  | Lvar id ->
      close_approx_var fenv cenv id
  | Lconst cst ->
      (match cst with
      | Const_base (Const_int n) -> (Jconst cst, Value_integer n)
      | Const_base (Const_char c) -> (Jconst cst, Value_integer (Char.code c))
      | Const_pointer n -> (Jconst cst, Value_constptr n)
      | _ -> (Jconst cst, Value_unknown))
  | Lfunction _ as func ->
      close_one_function fenv cenv (Ident.create "fun") func
  | Lapply (func, args, _) ->
      let nargs = List.length args in
      (match close fenv cenv func, close_list fenv cenv args with
      | (jfunc, Value_closure (fundesc, approx_res)),
         [Jprim (Pmakeblock _, jargs, _)]
         when List.length jargs = - fundesc.fun_arity ->
           let app = direct_apply fundesc func jfunc jargs in
           (app, strengthen_approx app approx_res)
      | (jfunc, Value_closure (fundesc, approx_res)), jargs
        when nargs = fundesc.fun_arity->
          let app = direct_apply fundesc func jfunc jargs in
          (app, strengthen_approx app approx_res)
      | (jfunc, Value_closure (fundesc, _)), jargs
        when fundesc.fun_arity > 0 && nargs > fundesc.fun_arity ->
          let first_args, rem_args = split_list fundesc.fun_arity jargs in
          Jgeneric_apply (direct_apply fundesc func jfunc first_args,
                          rem_args,
                          Debuginfo.none),
          Value_unknown
      | (jfunc, _), jargs ->
          Jgeneric_apply (jfunc, jargs, Debuginfo.none), Value_unknown)
  | Lsend (kind, meth, obj, params) ->
      let jmeth, _ = close fenv cenv meth in
      let jobj, _ = close fenv cenv obj in
      (Jsend (kind, jmeth, jobj, (close_list fenv cenv params), Debuginfo.none)), Value_unknown
  | Llet (str, id, lam, body) ->
      let jlam, alam = close_named fenv cenv id lam in
      (match str, alam with
      | Variable, _ ->
          let jbody, abody = close fenv cenv body in
          (Jlet (id, jlam, jbody)), abody
      | _, (Value_integer _ | Value_constptr _)
        when str = Alias || is_pure lam ->
          close (Tbl.add id alam fenv) cenv body
      | _ ->
          let jbody, abody = close (Tbl.add id alam fenv) cenv body in
          (Jlet (id, jlam, jbody)), abody)
  | Lletrec (defs, body) ->
      if List.for_all (function | _, Lfunction _ -> true | _ -> false) defs then begin
        let clos, infos = close_functions fenv cenv defs in
        let clos_ident = Ident.create "clos" in
        let fenv_body = List.fold_right
            (fun (id, _, approx) fenv -> Tbl.add id approx fenv)
            infos
            fenv in
        let jbody, approx = close fenv_body cenv body in
        let sb = List.fold_right
            (fun (id, pos, _) sb ->
              Tbl.add id (Joffset (Jvar clos_ident, pos)) sb)
            infos
            Tbl.empty in
        Jlet (clos_ident, clos, substitute sb jbody), approx
      end else begin
        let rec clos_defs = function
          | [] -> [], fenv
          | (id, lam) :: tail ->
              let jdefs, fenv_body = clos_defs tail in
              let jlam, approx = close fenv cenv lam in
              ((id, jlam) :: jdefs, Tbl.add id approx fenv_body) in
        let jdefs, fenv_body = clos_defs defs in
        let jbody, approx = close fenv_body cenv body in
        Jletrec (jdefs, jbody), approx
      end
  | Lprim (Pgetglobal id, []) as lam ->
      check_constant_result
        lam
        (getglobal id)
        (Jcompilenv.global_approx id)
  | Lprim (Pmakeblock (_, mut) as prim, lams) ->
      let jlams, approxs = List.split (List.map (close fenv cenv) lams) in
      (Jprim (prim, jlams, Debuginfo.none)),
      (match mut with
      | Immutable -> Value_tuple (Array.of_list approxs)
      | Mutable -> Value_unknown)
  | Lprim (Pfield n, [lam]) ->
      let jlam, approx = close fenv cenv lam in
      let fieldapprox = match approx with
      | Value_tuple a when n < Array.length a -> a.(n)
      | _ -> Value_unknown in
      check_constant_result lam (Jprim (Pfield n, [jlam], Debuginfo.none)) fieldapprox
  | Lprim (Psetfield (n, _), [Lprim (Pgetglobal id, []); lam]) ->
      let jlam, approx = close fenv cenv lam in
      (!global_approx).(n) <- approx;
      (Jprim (Psetfield (n, false), [getglobal id; jlam], Debuginfo.none)),
      Value_unknown
  | Lprim (Praise, [Levent (arg, ev)]) ->
      let jlam, _ = close fenv cenv arg in
      Jprim (Praise, [jlam], Debuginfo.from_raise ev), Value_unknown
  | Lprim (p, args) ->
      simplif_prim p (close_list_approx fenv cenv args) Debuginfo.none
  | Lswitch (arg, sw) ->
      let jarg, _ = close fenv cenv arg in
      let const_index, const_actions =
        close_switch fenv cenv sw.sw_consts sw.sw_numconsts sw.sw_failaction
      and block_index, block_actions =
        close_switch fenv cenv sw.sw_blocks sw.sw_numblocks sw.sw_failaction in
      (Jswitch (jarg, { js_index_consts = const_index;
                        js_actions_consts = const_actions;
                        js_index_blocks = block_index;
                        js_actions_blocks = block_actions})),
      Value_unknown
  | Lstaticraise (i, args) ->
      (Jstaticfail (i, (close_list fenv cenv args))), Value_unknown
  | Lstaticcatch (body, (i, vars), handler) ->
      let jbody, _ = close fenv cenv body in
      let jhandler, _ = close fenv cenv handler in
      (Jstaticcatch (i, vars, jbody, jhandler)), Value_unknown
  | Ltrywith (body, id, handler) ->
      let jbody, _ = close fenv cenv body in
      let jhandler, _ = close fenv cenv handler in
      (Jtrywith (jbody, id, jhandler)), Value_unknown
  | Lifthenelse (arg, ifso, ifno) ->
      (match close fenv cenv arg with
      | jarg, Value_constptr n ->
          sequence_constant_expr
            arg
            jarg
            (close fenv cenv (if n = 0 then ifno else ifso))
      | jarg, _ ->
          let jifso, _ = close fenv cenv ifso in
          let jifno, _ = close fenv cenv ifno in
          (Jifthenelse (jarg, jifso, jifno)), Value_unknown)
  | Lsequence (lam1, lam2) ->
      let jlam1, _ = close fenv cenv lam1 in
      let jlam2, approx = close fenv cenv lam2 in
      (Jsequence (jlam1, jlam2)), approx
  | Lwhile (cond, body) ->
      let jcond, _ = close fenv cenv cond in
      let jbody, _ = close fenv cenv body in
      (Jwhile (jcond, jbody)), Value_unknown
  | Lfor (id, lo, hi, dir, body) ->
      let jlo, _ = close fenv cenv lo in
      let jhi, _ = close fenv cenv hi in
      let jbody, _ = close fenv cenv body in
      (Jfor (id, jlo, jhi, dir, jbody)), Value_unknown
  | Lassign (id, lam) ->
      let jlam, _ = close fenv cenv lam in
      (Jassign (id, jlam)), Value_unknown
  | Levent (lam, ev) ->
      let jlam, approx = close fenv cenv lam in
      (add_debug_info ev jlam), approx
  | Lifused _ -> assert false
and close_list fenv cenv = function
  | [] -> []
  | lam :: tail ->
      let jlam, _ = close fenv cenv lam in
      jlam :: (close_list fenv cenv tail)
and close_list_approx fenv cenv = function
  | [] -> ([], [])
  | lam :: lams ->
      let jlam, approx = close fenv cenv lam in
      let jlams, approxs = close_list_approx fenv cenv lams in
      (jlam :: jlams, approx :: approxs)
and close_named fenv cenv id = function
  | (Lfunction _) as func -> close_one_function fenv cenv id func
  | lam -> close fenv cenv lam
and close_functions fenv cenv defs =
  incr depth;
  let initially_closed = !depth < excessive_depth in
  let fv = IdentSet.elements (free_variables (Lletrec (defs, lambda_unit))) in
  let uncurried_defs = List.map
      (function
        | (id, Lfunction (kind, params, body)) ->
            let label = Jcompilenv.make_symbol (Some (Ident.unique_name id)) in
            let arity = List.length params in
            let class_name = !Jclflags.java_package
              ^ "." ^ (Jcompilenv.current_unit_name ()) in
            let desc = { fun_label = class_name, label;
                         fun_arity = (if kind = Tupled then -arity else arity);
                         fun_closed = initially_closed;
                         fun_inline = None; } in
            (id, params, body, desc)
        | _ -> Misc.fatal_error "Jlambda.close_functions")
      defs in
  let fenv_rec = List.fold_right
      (fun (id, _, _, desc) fenv ->
        Tbl.add id (Value_closure (desc, Value_unknown)) fenv)
      uncurried_defs
      fenv in
  let env_pos = ref ~-1 in
  let clos_offsets = List.map
      (fun (_, _, _, desc) ->
        let pos = !env_pos + 1 in
        env_pos := !env_pos + 1 + (if desc.fun_arity <> 1 then 3 else 2);
        pos)
      uncurried_defs in
  let fv_pos = !env_pos in
  let useless_env = ref initially_closed in
  let clos_fundef (id, params, body, desc) env_pos =
    let env_param = Ident.create "env" in
    let cenv_fv = build_closure_env env_param (fv_pos - env_pos) fv in
    let cenv_body = List.fold_right2
        (fun (id, _, _, _) pos env ->
          Tbl.add id (Joffset (Jvar env_param, pos - env_pos)) env)
        uncurried_defs
        clos_offsets
        cenv_fv in
    let jbody, approx = close fenv_rec cenv_body body in
    if !useless_env && occurs_var env_param jbody then useless_env := false;
    let fun_params = if !useless_env then params else params @ [env_param] in
    ((desc.fun_label, desc.fun_arity, fun_params, jbody),
     (id, env_pos, Value_closure (desc, approx))) in
  let clos_info_list =
    if initially_closed then begin
      let cl = List.map2 clos_fundef uncurried_defs clos_offsets in
      if !useless_env then
        cl
      else begin
        List.iter
          (fun (_, _, _, desc) -> desc.fun_closed <- false)
          uncurried_defs;
        List.map2 clos_fundef uncurried_defs clos_offsets
      end
    end else
      List.map2 clos_fundef uncurried_defs clos_offsets in
  decr depth;
  let close, infos = List.split clos_info_list in
  Jclosure (close, List.map (close_var fenv cenv) fv), infos
and close_one_function fenv cenv id func =
  match (close_functions fenv cenv [id, func]) with
  | (Jclosure ([_, _, params, body], _) as clos),
    [_, _, (Value_closure (fundesc, _) as approx)] ->
      if lambda_smaller body (!Clflags.inline_threshold + List.length params)
      then fundesc.fun_inline <- Some (params, body);
      clos, approx
  | _ -> Misc.fatal_error "Jlambda.close_one_function"
and close_switch fenv cenv cases num_keys default =
  let index = Array.create num_keys 0
  and store = Switch.mk_store Lambda.same in
  (match default with
  | Some def when List.length cases < num_keys ->
      ignore (store.Switch.act_store def)
  | _ -> ());
  List.iter
    (fun (key, lam) ->
      index.(key) <- store.Switch.act_store lam)
    cases;
  let actions =
    Array.map
      (fun lam -> fst (close fenv cenv lam))
      (store.Switch.act_get ()) in
  match actions with
  | [| |] -> [| |], [| |]
  | _ -> index, actions


let jlambda_of_lambda size lam =
  depth := 0;
  global_approx := Array.create size Value_unknown;
  Jcompilenv.set_global_approx (Value_tuple !global_approx);
  let jlam, _ = close Tbl.empty Tbl.empty lam in
  global_approx := [||];
  jlam
