(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2010 Xavier Clerc.
 * Original file (asmcomp/compilenv.mli in the Objective Caml source
 * distribution) is Copyright (C) INRIA.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://qt.nokia.com/doc/4.0/qpl.html>.
 *)

(** Compilation environments. *)

type error =
  | Not_a_unit_info of string
  | Corrupted_unit_info of string
  | Illegal_renaming of string* string

exception Error of error


type unit_infos =
    { mutable ui_name : string;
      mutable ui_javaname : string;
      mutable ui_additional_classes : string list;
      mutable ui_symbol : string;
      mutable ui_defines : string list;
      mutable ui_imports_cmi : (string * Digest.t) list;
      mutable ui_imports_cmj : (string * Digest.t) list;
      mutable ui_approx : Jlambda.value_approximation; }
(** Structure for .cmj files, that are the Java conterpart of .cmx native
   files. *)

val cmj_not_found_crc : Digest.t
(** Digest used to signal that a .cmj file is missing. *)

type library_infos =
    { lib_units : (unit_infos * Digest.t) list;
      lib_ccobjs : string list; }
(** Structure for .cmja files, that are the Java counterpart of .cmxa native
   files. *)

val reset : ?packname:string -> string -> unit
(** Resets the current environment. The mandatory paremeter is the name of
   the unit being compiled while the optional parameter is its package name
   (this package name is for [-for-pack] handling, that is package on the
   OCaml side, not on the Java side). *)

val current_unit_infos : unit -> unit_infos
(** Returns the infos for the unit being compiled. *)

val current_unit_name : unit -> string
(** Returns the name of the unit being compiled. *)

val read_unit_info : string -> unit_infos * Digest.t
(** [read_unit_info file] reads both infos and digest from [file]
   (should be a .cmj file). *)

val write_unit_info : unit_infos -> string -> unit
(** [write_unit_info infos file] writes both [infos] and its digest to
   [file]. *)

val save_unit_info : string -> unit
(** Save the infos for the unit being compiled to the passed file. *)

val cache_unit_info : unit_infos -> unit
(** Adds the passed infos to the cache. There will thus be no subsequent need
   to read them from the corresponding file. *)

val make_symbol : ?unitname:string -> string option -> string
(** [make_symbol u id] constructs a symbol for unit [u] (defaulting to the
   current unit), [id] being the base name of the symbol.
   If [id] is [None] then the symbol corresponding the the unit as a whole
   is returned. *)

val symbol_for_global : Ident.t -> string
(** Returns the symbol corresponding to the passed global identifier. *)

val global_approx : Ident.t -> Jlambda.value_approximation
(** Returns the approximation corresponding to the passed global identifier. *)

val set_global_approx : Jlambda.value_approximation -> unit
(** Sets the approximation for the unit being compiled. *)

val record_global_approx_toplevel : unit -> unit
(** Records the current approximation for the current toplevel phrase. *)

val read_library_info : string -> library_infos
(** Reads the library info for the passed filename. *)

val report_error : Format.formatter -> error -> unit
(** Pretty-prints an error. *)
