(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://qt.nokia.com/doc/4.0/qpl.html>.
 *)

open Format
open Macroinstr


let rec expr ppf = function
  | Mconst cst ->
      Printlambda.structured_constant ppf cst
  | Moffset (e, n) ->
      fprintf ppf "@[<2>(offset@ %a@ %d)@]"
        expr e
        n
  | Mdynamicoffset (e1, e2) ->
      fprintf ppf "@[<2>(dynamicoffset@ %a@ %a)@]"
        expr e1
        expr e2
  | Mcreateglobal (s, n) ->
      fprintf ppf "@[<2>(createglobal@ %s@ %d)@]"
        s
        n
  | Mclosure (_, funcs, vars) ->
      let fun_decls ppf l =
        List.iter
          (fun ((cls, func), arity) ->
            fprintf ppf "@ (%s@ %s@ %d)" cls func arity)
          l in
      fprintf ppf "@[<2>(closure%a%a)@]"
        fun_decls funcs
        exprs vars
  | Mreadlocal n ->
      fprintf ppf "@[<2>(readlocal@ %d)@]" n
  | Mreadlocalint n ->
      fprintf ppf "@[<2>(readlocalint@ %d)@]" n
  | Mwritelocal (n, e) ->
      fprintf ppf "@[<2>(writelocal@ %d@ %a)@]"
        n
        expr e
  | Mwritelocalint (n, e) ->
      fprintf ppf "@[<2>(writelocalint@ %d@ %a)@]"
        n
        expr e
  | Mcall (call, (cls, func), l, _) ->
      fprintf ppf "@[<2>(call@ %s@ %s@ %s%a)@]"
        (if call = Primitive then "prim" else "native")
        cls
        func
        exprs l
  | Mprim (prim, l, _) ->
      fprintf ppf "@[<2>(prim@ %a%a)@]"
        Printjlambda.primitive prim
        exprs l
  | Mapply (l, _) ->
      fprintf ppf "@[<2>(apply%a)@]" exprs l
  | Msequence (e1, e2) ->
      fprintf ppf "@[<2>(seq@ %a@ %a)@]"
        expr e1
        sequence e2
  | Mifthenelse (cond, ifso, ifno) ->
      fprintf ppf "@[<2>(if@ %a@ %a@ %a)@]"
        expr cond
        expr ifso
        expr ifno
  | Mwhile (cond, body) ->
      fprintf ppf "@[<2>(while@ %a@ %a)@]"
        expr cond
        expr body
  | Mfor (n, dir, n', e) ->
      fprintf ppf "@[<2>(for@ %d@ %s@ %d@ %a)@]"
        n
        (if dir = Asttypes.Upto then "to" else "downto")
        n'
        expr e
  | Munit ->
      fprintf ppf "@[<2>(unit)@]"
  | Mpoptolocal n ->
      fprintf ppf "@[<2>(poptolocal@ %d)@]" n
  | Mpop ->
      fprintf ppf "@[<2>(pop)@]"
  | Mnop ->
      fprintf ppf "@[<2>(nop)@]"
  | Mtrywith (body, n, handler, _) ->
      fprintf ppf "@[<2>(try@ %a@;<1 -1>with %d@ %a)@]"
        expr body
        n
        expr handler
  | Mstaticfail (nfail, l) ->
      fprintf ppf "@[<2>(staticfail@ %d%a)@]"
        nfail
        exprs l
  | Mstaticcatch (nfail, body, n, handler) ->
      fprintf ppf "@[(catch@ %a@;<1 -1>with %d@ %d %a)@]"
        expr body
        nfail
        n
        expr handler
  | Mswitch (e, index, acts, default) ->
      let couples ppf l =
        let spc = ref false in
        Array.iter
          (fun (n, e) ->
            if !spc then fprintf ppf "@ " else spc := true;
            fprintf ppf "@[<hv 1>case %i:@ %a@]" n expr e)
          l in
      fprintf ppf "@[<1>(switch %a@ @[<v 0>%a@]@ default: %a)@]"
        expr e
        couples (Array.mapi (fun i x -> i, acts.(x)) index)
        expr default
  | Mint e ->
      fprintf ppf "@[<2>(int@ %a)@]" expr e
  | Mtag e ->
      fprintf ppf "@[<2>(tag@ %a)@]" expr e
and exprs ppf l =
  List.iter (fun e -> fprintf ppf "@ %a" expr e) l
and sequence ppf = function
  | Msequence (e1, e2) -> fprintf ppf "%a@ %a" sequence e1 sequence e2
  | e -> expr ppf e

let minstr ppf fundecls =
  let idents ppf l =
    List.iter (fun id -> fprintf ppf "@ %a" Ident.print id) l in
  let spc = ref false in
  List.iter
    (fun fd ->
      if !spc then fprintf ppf "@ " else spc := true;
      fprintf ppf "@[(function@ %s%a@ %d@ %d@ %a)@]"
        fd.fun_name
        idents fd.fun_args
        fd.fun_max_locals
        fd.fun_garbage_index
        expr fd.fun_body)
    fundecls
