(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2010 Xavier Clerc.
 * Original file (toplevel/toploop.ml in the Objective Caml source
 * distribution) is Copyright (C) INRIA.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://qt.nokia.com/doc/4.0/qpl.html>.
 *)

open Config
open Parsetree


type evaluation_outcome = Result of Obj.t | Exception of exn

let load_lambda lam =
  let slam = Simplif.simplify_lambda lam in
  let (init_code, fun_code) = Bytegen.compile_phrase slam in
  let (code, code_size, reloc) = Emitcode.to_memory init_code fun_code in
  let can_free = (fun_code = []) in
  let initial_symtable = Symtable.current_state() in
  Symtable.patch_object code reloc;
  Symtable.check_global_initialized reloc;
  Symtable.update_global_table();
  try
    let retval = (Meta.reify_bytecode code code_size) () in
    if can_free then begin
      Meta.static_release_bytecode code code_size;
      Meta.static_free code;
    end;
    Result retval
  with x ->
    if can_free then begin
      Meta.static_release_bytecode code code_size;
      Meta.static_free code;
    end;
    Symtable.restore_state initial_symtable;
    Exception x

let toplevel_env = ref Env.empty

let execute_phrase phr =
  match phr with
  | Ptop_def sstr ->
      let oldenv = !toplevel_env in
      Typecore.reset_delayed_checks ();
      let (str, _, newenv) = Typemod.type_structure oldenv sstr Location.none in
      Typecore.force_delayed_checks ();
      let lam = Translmod.transl_toplevel_definition str in
      begin try
        toplevel_env := newenv;
        let res = load_lambda lam in
        match res with
        | Result v -> Obj.magic v
        | Exception exn ->
            toplevel_env := oldenv;
            if exn = Out_of_memory then Gc.full_major();
            raise exn
      with x ->
        toplevel_env := oldenv; raise x
      end
  | _ -> 
      false

let first_line = ref true
let got_eof = ref false

let read_input_default buffer len =
  let i = ref 0 in
  try
    while true do
      if !i >= len then raise Exit;
      let c = input_char stdin in
      buffer.[!i] <- c;
      incr i;
      if c = '\n' then raise Exit;
    done;
    (!i, false)
  with
  | End_of_file ->
      (!i, true)
  | Exit ->
      (!i, false)

let refill_lexbuf buffer len =
  if !got_eof then (got_eof := false; 0) else begin
    first_line := false;
    let (len, eof) = read_input_default buffer len in
    if eof then begin
      Location.echo_eof ();
      if len > 0 then got_eof := true;
      len
    end else
      len
  end

let _ =
  Sys.interactive := true;
  let crc_intfs = Symtable.init_toplevel() in
  Compile.init_path();
  ignore (Topdirs.load_file Format.std_formatter "stdlib.cma");
  List.iter
    (fun (name, crc) ->
      Consistbl.set Env.crc_units name crc Sys.executable_name)
    crc_intfs

let set_paths () =
  load_path := !load_path @ [Filename.concat Config.standard_library "camlp4"];
  load_path := "" :: (List.rev !Clflags.include_dirs @ !load_path);
  Dll.add_path !load_path

let initialize_toplevel_env () =
  toplevel_env := Compile.initial_env()

let eval s =
  (* init *)
  let lb = Lexing.from_string s in
  Location.input_name := "";
  Location.input_lexbuf := Some lb;
  (* eval *)
  try
    ignore (Btype.snapshot ());
    first_line := true;
    let phrases = Parse.use_file lb in
    List.fold_left
      (fun _ phrase ->
        execute_phrase phrase)
      (Obj.magic ())
      phrases
  with
  | e ->
      let buf = Buffer.create 32 in
      let fmt = Format.formatter_of_buffer buf in
      (try Errors.report_error fmt e with _ -> failwith "unable to compile script");
      Format.pp_print_flush fmt ();
      failwith (Buffer.contents buf)

let load s =
  ignore (Topdirs.load_file Format.err_formatter s)

let directory s =
  Topdirs.dir_directory s

let init () =
  initialize_toplevel_env ();
  let lb = Lexing.from_function refill_lexbuf in
  Location.input_name := "";
  Location.input_lexbuf := Some lb;
  Callback.register "javax.script.eval" eval;
  Callback.register "javax.script.directory" directory;
  Callback.register "javax.script.load" load
